# @PydevCodeAnalysisIgnore
import pandas as pd
from numpy.core.numeric import nan
from .messages import Message
from .errors import Error
from .db import Db
from .charts import Plot
from .maps import Map
from .data import Df
from .report import Report

__version__ = "0.4.25"


class Ds(Db, Df, Plot, Map, Report, Error, Message):
    """
    Main class
    """

    def __init__(self, df=None, db=None, nbload_libs=True):
        """
        Initialize with an empty dataframe
        """
        self.msg = Message()
        self.version = __version__
        self.df = df
        self.db = db
        self.x = None
        self.y = None
        self.chart_obj = None
        self.chart_opts = dict(width=880)
        self.chart_style = {}
        self.engine = "bokeh"
        self.label = None
        self.backup_df = None
        self.autoprint = False
        self.errors_handling = "exceptions"
        self.notebook = False
        self.header = self._header
        self.footer = self._footer
        self.reports = []
        self.report_engines = []
        self.start_time = None
        self.influx_cli = None
        self.datapath = None
        self.report_path = None
        self.static_path = None
        self.quiet = False
        self.nan = nan
        self.color_index = 0
        self.dsmap = None
        self.altair_encode = {}

        if self._isnotebook() is True and nbload_libs is True:
            self.notebook = True
            import holoviews as hv
            hv.extension("bokeh")
            import altair as alt
            alt.renderers.enable('notebook')

    def __repr__(self):
        num = 0
        if self.df is not None:
            num = len(self.df.index)
        msg = "<DataSwim object | " + str(num) + " rows>"
        return msg

    def new_(self, df=pd.DataFrame(), db=None, quiet=False,
             nbload_libs=True):
        """
        Returns a new DataSwim instance from a dataframe
        """
        ds2 = Ds(df, db, nbload_libs)
        if quiet is False:
            self.ok("A new instance was created")
        return ds2

    def _isnotebook(self):
        try:
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True
            else:
                return False
        except NameError:
            return False



from ...errors import Error
from ...messages import Message


class Columns(Error, Message):
    """
    Class to transform the dataframe's columns
    """

    def rename(self, source_col: str, dest_col: str):
        """
        Renames a column in the main dataframe

        :param source_col: name of the column to rename
        :type source_col: str
        :param dest_col: new name of the column
        :type dest_col: str

        :example: ``ds.rename("Col 1", "New col")``
        """
        try:
            self.df = self.df.rename(columns={source_col: dest_col})
        except Exception as e:
            self.err(e, self.rename, "Can not rename column")
            return
        self.ok("Column", source_col, "renamed")

    def add(self, col: str, value):
        """
        Add a column with default values

        :param col: column name
        :type col: str
        :param value: column value
        :type value: any

        :example: ``ds.add("Col 4", 0)``
        """
        try:
            self.df[col] = value
        except Exception as e:
            self.err(e, self.add, "Can not add column")

    def keep_(self, *cols) -> "Ds":
        """
        Returns a dataswim instance with a dataframe limited
        to some columns

        :param cols: names of the columns
        :type cols: str
        :return: a dataswim instance
        :rtype: Ds

        :example: ``ds2 = ds.keep_("Col 1", "Col 2")``
        """
        try:
            ds2 = self._duplicate_(self.df[list(cols)])
        except Exception as e:
            self.err(e, "Can not remove colums")
            return
        self.ok("Columns", " ,".join(cols), "kept")
        return ds2

    def keep(self, *cols):
        """
        Limit the dataframe to some columns

        :param cols: names of the columns
        :type cols: str

        :example: ``ds.keep("Col 1", "Col 2")``
        """
        try:
            self.df = self.df[list(cols)]
        except Exception as e:
            self.err(e, "Can not remove colums")
            return
        self.ok("Setting dataframe to columns", " ".join(cols))

    def drop(self, *cols):
        """
        Drops columns from the main dataframe

        :param cols: names of the columns
        :type cols: str

        :example: ``ds.drop("Col 1", "Col 2")``
        """
        try:
            index = self.df.columns.values
            for col in cols:
                if col not in index:
                    self.warning("Column", col, "not found. Aborting")
                    return
                self.df = self.df.drop(col, axis=1)
        except Exception as e:
            self.err(e, self.drop, "Can not drop column")

    def exclude(self, col: str, val):
        """
        Delete rows based on value

        :param col: column name
        :type col: str
        :param val: value to delete
        :type val: any

        :example: ``ds.exclude("Col 1", "value")``
        """
        try:
            self.df = self.df[self.df[col] != val]
        except Exception as e:
            self.err(e, "Can not exclude rows based on value " + str(val))

    def copycol(self, origin_col: str, dest_col: str):
        """
        Copy a columns values in another column

        :param origin_col: name of the column to copy
        :type origin_col: str
        :param dest_col: name of the new column
        :type dest_col: str

        :example: ``ds.copy("col 1", "New col")``
        """
        try:
            self.df[dest_col] = self.df[[origin_col]]
        except Exception as e:
            self.err(e, self.copy_col, "Can not copy column")

    def indexcol(self, col: str):
        """
        Add a column from the index

        :param col: name of the new column
        :type col: str

        :example: ``ds.index_col("New col")``
        """
        try:
            self.df[col] = self.df.index.values
        except Exception as e:
            self.err(e)
            return
        self.ok("Column", col, "added from the index")



# -*- coding: utf-8 -*
from altair import Chart


class Altair():
    """
    A class to handle charts with the Altair library
    """

    def __init__(self, df=None):
        """
        Initialize
        """
        self.df = df
        self.altair_encode = {}

    def aenc(self, key, value):
        """
        Add an entry to the altair encoding dict
        """
        self.altair_encode[key] = value

    def raenc(self, key):
        """
        Remove an entry from the altair encoding dict
        """
        if key in self.altair_encode:
            del self.altair_encode[key]
        else:
            self.warning("Key " + key + " not found in Altair encoding dict")

    def raencs(self):
        """
        Reset the altair encoding dict
        """
        self.altair_encode = {}

    def _altair_line_num_(self, xfield, yfield, opts, style, encode):
        """
        Get a line + text number chart
        """
        try:
            c = self._altair_chart_num_("line", xfield,
                                           yfield, opts, style, encode)
        except Exception as e:
            self.err(e, "Can not draw a line num chart")
            return
        return c

    def _altair_bar_num_(self, xfield, yfield, opts, style, encode):
        """
        Get a bar + text number chart
        """
        try:
            c = self._altair_chart_num_("bar", xfield,
                                           yfield, opts, style, encode)
        except Exception as e:
            self.err(e, "Can not draw a bar num chart")
            return
        return c
    
    def _altair_point_num_(self, xfield, yfield, opts, style, encode):
        """
        Get a point + text number chart
        """
        try:
            c = self._altair_chart_num_("point", xfield,
                                           yfield, opts, style, encode)
        except Exception as e:
            self.err(e, "Can not draw a line num chart")
            return
        return c

    def _altair_chart_num_(self, chart_type, xfield, yfield, opts, style2, encode):
        """
        Get a chart + text number chart
        """
        style = {**style2}
        text_color = "grey"
        if "text_color" in style:
            text_color = style["text_color"]
            del style["text_color"]
            if "text_color" in self.chart_style:
                del self.chart_style["text_color"]
        if chart_type == "line":
            c = Chart(self.df).mark_line(**style).encode(x=xfield, \
                                                y=yfield, **encode).properties(**opts)
        elif chart_type == "bar":
            c = Chart(self.df).mark_bar(**style).encode(x=xfield, \
                                                y=yfield, **encode).properties(**opts)
        elif chart_type == "point":
            c = Chart(self.df).mark_point(**style).encode(x=xfield, \
                                                y=yfield, **encode).properties(**opts)
        encoder = encode
        if "text" not in encoder:
            encoder["text"] = yfield
        if "align" not in style:
            style["align"] = "center"
        if "dy" not in style:
            style["dy"] = -5
        if "dx" not in style and chart_type != "bar":
            style["dx"] = 8
        if "size" in style:
            del style["size"]
        style["color"] = text_color
        df2 = self.df.replace({yfield.split(":")[0]: {0: self.nan}})
        num = Chart(df2).mark_text(**style).encode(x=xfield, \
                                            y=yfield, **encoder).properties(**opts)
        return c + num

    def _altair_hline_(self, xfield, yfield, opts, style, encode):
        """
        Get a mean line chart
        """
        try:
            rawy = yfield
            if ":" in yfield:
                rawy = yfield.split(":")[0]
            mean = self.df[rawy].mean()
            l = []
            i = 0
            while i < len(self.df[rawy]):
                l.append(mean)
                i += 1
            self.df["Mean"] = l
            chart = Chart(self.df).mark_line(**style).encode(x=xfield, \
                                            y="Mean", **encode).properties(**opts)
            self.drop("Mean")
            return chart
        except Exception as e:
            self.err(e, "Can not draw mean line chart")

    def _get_altair_chart(self, xfield, yfield, chart_type,
                          label, opts={}, style={}, **kwargs):
        """
        Get an Altair chart object
        """
        encode = self.altair_encode
        chart = None
        if chart_type == "bar":
            chart = Chart(self.df).mark_bar(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "circle":
            chart = Chart(self.df).mark_circle(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "line":
            chart = Chart(self.df).mark_line(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "hline":
            chart = self._altair_hline_(xfield, yfield, opts, style, encode)
        elif chart_type == "line_num":
            chart = self._altair_line_num_(xfield, yfield, opts, style, encode)
        elif chart_type == "bar_num":
            chart = self._altair_bar_num_(xfield, yfield, opts, style, encode)
        elif chart_type == "point_num":
            chart = self._altair_point_num_(xfield, yfield, opts, style, encode)
        elif chart_type == "point":
            chart = Chart(self.df).mark_point(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "area":
            chart = Chart(self.df).mark_area(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "heatmap":
            chart = Chart(self.df).mark_rect(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "text":
            chart = Chart(self.df).mark_text(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "square":
            chart = Chart(self.df).mark_square(**style).encode(x=xfield, \
                                                y=yfield, **encode).properties(**opts)
        elif chart_type == "tick":
            chart = Chart(self.df).mark_tick(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        elif chart_type == "rule":
            chart = Chart(self.df).mark_rule(**style).encode(x=xfield, \
                                            y=yfield, **encode).properties(**opts)
        return chart

    def altair_header_(self):
        """
        Returns html script tags for Altair
        """
        header = """
        <script src="https://cdn.jsdelivr.net/npm/vega@3"></script>
        <script src="https://cdn.jsdelivr.net/npm/vega-lite@2"></script>
        <script src="https://cdn.jsdelivr.net/npm/vega-embed@3"></script>
        <style>.vega-actions {display:none}</style>
        """
        return header

    def _get_altair_html_(self, chart_obj, slug):
        """
        Get html for an Altair chart
        """
        try:
            json_data = chart_obj.to_json(indent=0)
        except Exception as e:
            self.err(e)
        html = '<div id="' + slug + '"></div>\n'
        html += '<script type="text/javascript">'
        html += 'var spec = ' + json_data.replace("\n", "") + ";"
        html += """
        var embed_opt = {"mode": "vega-lite"};
        function showError(altel, error){
            altel.innerHTML = ('<div class="error">'
                            + '<p>JavaScript Error: ' + error.message + '</p>'
                            + "<p>This usually means there's a typo in your chart specification. "
                            + "See the javascript console for the full traceback.</p>"
                            + '</div>');
            throw error;
        };\n"""
        html += "const el_" + slug + " = document.getElementById('" + slug + "');"
        html += "vegaEmbed('#" + slug + "', spec, embed_opt)"
        html += ".catch(error => showError(el_" + slug + ", error));"
        html += '</script>'
        return html

    def _set_altair_engine(self):
        """
        Set the current chart engine to Altair
        """
        if self.engine != "altair":
            self.engine = "altair"
            self.info("Switching to the Altair engine to draw this chart")



import pandas as pd
from pandas.testing import assert_frame_equal
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()

ds.errs_traceback = False


class TestDsDataInit(BaseDsTest):

    def test_clone_duplicate_(self):
        ds.df = self.df
        ds2 = ds.clone_()
        assert_frame_equal(ds2.df, ds.df)
        ds2 = ds.new_()
        # error clone
        ds.df = None
        ds2 = ds.clone_()
        self.assertRaises(AttributeError)
        ds.df = "wrong"
        ds2 = ds.clone_()
        self.assertRaises(AttributeError)
        ds2 = ds._duplicate_(self.df, "db", False)

    def test_load_json(self):
        data = self.df.to_json()
        ds.load_json(data)
        ds.load_json("wrong")
        self.assertRaises(ValueError)

    def test_load_csv(self):
        ds.load_csv(self.path+"/data/fixtures/data.csv")
        ds.datapath = self.path+"/data"
        ds.load_csv("data.csv")
        ds.load_csv(None)

    def test_load_excel(self):
        ds.load_excel(self.path+"/data/fixtures/data.xls")
        ds.load_excel("wrong")
        ds.load_excel(self.path+"/data/fixtures/empty.xls")

    def test_load_h5(self):
        ds.load_h5(self.path + "/data/fixtures/data.h5")
        ds.load_h5("wrong")

    def test_dateparser(self):
        dp = ds.dateparser()
        dp(["12/01/2012", "13/01/2012"])

    def test_backup_restore(self):
        ds.df = self.df
        ds.backup()
        ds.df = pd.DataFrame({"1": 1}, ["1"])
        ds.restore()
        assert_frame_equal(ds.df, self.df)
        # errors
        ds.df = "wrong"
        ds.backup()
        self.assertRaises(AttributeError)
        ds.df = self.df
        ds.backup_df = "wrong"
        ds.restore()
        self.assertRaises(AttributeError)
        ds.backup_df = None
        ds.restore()
        self.assertRaises(AttributeError)



import pandas as pd
from ..copy import Copy


class Values(Copy):
    """
    A class to transform the dataframe's values
    """

    def sum_(self, col: str) -> float:
        """
        Returns the sum of all values in a column

        :param col: column name
        :type col: str
        :return: sum of all the column values
        :rtype: float

        :example: ``sum = ds.sum_("Col 1")``
        """
        try:
            df = self.df[col]
            num = float(df.sum())
            return num
        except Exception as e:
            self.err(e, "Can not sum data on column " + str(col))

    def dropr(self, *rows):
        """
        Drops some rows from the main dataframe

        :param rows: rows names
        :type rows: list of ints

        :example: ``ds.drop_rows([0, 2])``
        """
        try:
            self.df = rows = list(rows)
            self.df.drop(rows)
        except Exception as e:
            self.err(e, self.dropr, "Can not drop rows")

    def append(self, vals: list, index=None):
        """
        Append a row to the main dataframe

        :param vals: list of the row values to add
        :type vals: list
        :param index: index key, defaults to None
        :param index: any, optional

        :example: ``ds.append([0, 2, 2, 3, 4])``
        """
        try:
            # self.df.append(vals)
            self.df.append(pd.DataFrame(columns=self.df.columns, data=[vals]))
        except Exception as e:
            self.err(e, self.append, "Can not append row")
            return
        self.ok("Row added to dataframe")

    def sort(self, col: str):
        """
        Sorts the main dataframe according to the given column

        :param col: column name
        :type col: str

        :example: ``ds.sort("Col 1")``
        """
        try:
            self.df = self.df.copy().sort_values(col)
        except Exception as e:
            self.err(e, "Can not sort the dataframe from column " +
                     str(col))

    def reverse(self):
        """
        Reverses the main dataframe order

        :example: ``ds.reverse()``
        """
        try:
            self.df = self.df.iloc[::-1]
        except Exception as e:
            self.err(e, "Can not reverse the dataframe")

    def apply(self, function: "function", *cols, axis=1, **kwargs):
        """
        Apply a function on columns values

        :param function: a function to apply to the columns
        :type function: function
        :param cols: columns names
        :type cols: name of columns
        :param axis: index (0) or column (1), default is 1
        :param kwargs: arguments for ``df.apply``
        :type kwargs: optional

        :example:
                        .. code-block:: python

                                def f(row):
                                        # add a new column with a value
                                        row["newcol"] = row["Col 1] + 1
                                        return row

                                ds.apply(f)

        """
        try:
            if len(cols) == 0:
                self.df = self.df.apply(function, axis=axis, **kwargs)
            else:
                cols = list(cols)
                self.df[cols] = self.df[cols].apply(function, **kwargs)
        except Exception as e:
            self.err(e, "Can not apply function")

    def pivot(self, index, **kwargs):
        """
        Pivots a dataframe
        """
        try:
            return pd.pivot_table(self.df, index=kwargs["index"],  **kwargs)
            """if df is None:
                self.err("Can not pivot table")
                return
            self.df = df"""
        except Exception as e:
            self.err(e, "Can not pivot dataframe")

    def trimquants(self, col: str, inf: float, sup: float):
        """
        Remove superior and inferior quantiles from the dataframe

        :param col: column name
        :type col: str
        :param inf: inferior quantile
        :type inf: float
        :param sup: superior quantile
        :type sup: float

        :example: ``ds.trimquants("Col 1", 0.01, 0.99)``
        """
        try:
            self.df = self._trimquants(col, inf, sup)
        except Exception as e:
            self.err(e, self.trimquants, "Can not trim quantiles")

    def trimsquants(self, col: str, sup: float):
        """
        Remove superior quantiles from the dataframe

        :param col: column name
        :type col: str
        :param sup: superior quantile
        :type sup: float

        :example: ``ds.trimsquants("Col 1", 0.99)``
        """
        try:
            self.df = self._trimquants(col, None, sup)
        except Exception as e:
            self.err(e, self.trimsquants, "Can not trim superior quantiles")

    def trimiquants(self, col: str, inf: float):
        """
        Remove superior and inferior quantiles from the dataframe

        :param col: column name
        :type col: str
        :param inf: inferior quantile
        :type inf: float

        :example: ``ds.trimiquants("Col 1", 0.05)``
        """
        try:
            self.df = self._trimquants(col, inf, None)
        except Exception as e:
            self.err(e, self.trimiquants, "Can not trim inferior quantiles")

    def _trimquants(self, col, inf, sup):
        try:
            ds2 = self._duplicate_()
            if inf is not None:
                qi = ds2.df[col].quantile(inf)
                ds2.df = ds2.df[ds2.df[col] > qi]
            if sup is not None:
                qs = ds2.df[col].quantile(sup)
                ds2.df = ds2.df[ds2.df[col] < qs]
        except Exception as e:
            self.err(e, self._trimquants, "Can not trim quantiles")
            return
        msg = "Removed values "
        if inf is not None:
            msg += "under " + str(qi)
        if sup is not None and inf is not None:
            msg += " and"
        if sup is not None:
            msg += "upper " + str(qs)
        self.ok(msg, "in column", col)
        return ds2.df



from goerr.colors import colors


class Search():
    """
    Class to handle searches in the dataframe
    """

    def contains(self, column, value):
        """
        Set the main dataframe instance to rows that contains a string
        value in a column
        """
        df = self.df[self.df[column].str.contains(value) == True]
        if df is None:
            self.err("Can not select contained data")
            return
        self.df = df

    def ncontains(self, column, value):
        """
        Set the main dataframe instance to rows that do not
        contains a string value in a column
        """
        df = self.df[self.df[column].str.contains(value) == False]
        if df is None:
            self.err("Can not select contained data")
            return
        self.df = df

    def exact(self, column, *values):
        """
        Sets the main dataframe to rows that has the exact string
        value in a column
        """
        df = self._exact(column, *values)
        if df is None:
            self.err("Can not select exact data")
        self.df = df

    def exact_(self, column, *values):
        """
        Returns a Dataswim instance with rows that has the exact string
        value in a column
        """
        df = self._exact(column, *values)
        if df is None:
            self.err("Can not select exact data")
        return self._duplicate_(df)

    def _exact(self, column, *values):
        try:
            df2 = self.df[column].isin(values)
            df = self.df[df2]
            return df
        except Exception as e:
            self.err(e, "Can not select exact data")



import pandas as pd
from goerr.colors import colors
from .copy import Copy


class View(Copy):
    """
    Class to view the data
    """

    def show(self, rows: int=5,
             dataframe: pd.DataFrame=None) -> pd.DataFrame:
        """
        Display info about the dataframe

        :param rows: number of rows to show, defaults to 5
        :param rows: int, optional
        :param dataframe: a pandas dataframe, defaults to None
        :param dataframe: pd.DataFrame, optional
        :return: a pandas dataframe
        :rtype: pd.DataFrame

        :example: ``ds.show()``
        """
        try:
            if dataframe is not None:
                df = dataframe
            else:
                df = self.df
            if df is None:
                self.warning("Dataframe is empty: nothing to show")
                return
            num = len(df.columns.values)
        except Exception as e:
            self.err(e, self.show, "Can not show dataframe")
            return
        f = list(df)
        fds = []
        for fi in f:
            fds.append(str(fi))
        fields = ", ".join(fds)
        num_rows = len(df.index)
        self.info("The dataframe has", colors.bold(num_rows), "rows and",
                  colors.bold(num), "columns:")
        print(fields)
        return df.head(rows)

    def one(self):
        """
        Shows one row of the dataframe and the field 
        names wiht count

        :return: a pandas dataframe
        :rtype: pd.DataFrame

        :example: ``ds.one()``
        """
        try:
            return self.show(1)
        except Exception as e:
            self.err(e, self.one, "Can not display dataframe")

    def tail(self, rows: int=5):
        """
        Returns the main dataframe's tail

        :param rows: number of rows to print, defaults to 5
        :param rows: int, optional
        :return: a pandas dataframe
        :rtype: pd.DataFrame

        :example: ``ds.tail()``
        """
        if self.df is None:
            self.warning("Dataframe is empty: no tail available")
        return self.df.tail(rows)

    def cols_(self) -> pd.DataFrame:
        """
        Returns a dataframe with columns info

        :return: a pandas dataframe
        :rtype: pd.DataFrame

        :example: ``ds.cols_()``
        """
        try:
            s = self.df.iloc[0]
            df = pd.DataFrame(s)
            df = df.rename(columns={0: "value"})

            def run(row):
                t = row[0]
                return type(t).__name__

            s = df.apply(run, axis=1)
            df = df.rename(columns={0: "value"})
            df["types"] = s
            return df
        except Exception as e:
            self.err(e)

    def describe_(self):
        """
        Return a description of the data

        :return: a pandas dataframe
        :rtype: pd.DataFrame

        :example: ``ds.describe()``
        """
        if self.df is None:
            self.warning("Dataframe is empty: nothing to describe")
            return
        return self.df.describe()

    def types_(self, col: str) -> pd.DataFrame:
        """
        Display types of values in a column

        :param col: column name
        :type col: str
        :return: a pandas dataframe
        :rtype: pd.DataFrame

        :example: ``ds.types_("Col 1")``
        """
        cols = self.df.columns.values
        all_types = {}
        for col in cols:
            local_types = []
            for i, val in self.df[col].iteritems():
                #print(i, val, type(val))
                t = type(val).__name__
                if t not in local_types:
                    local_types.append(t)
            all_types[col] = (local_types, i)
        df = pd.DataFrame(all_types, index=["type", "num"])
        return df



# -*- coding: utf-8 -*-
import datetime
import pandas as pd
import numpy as np
from numpy import nan
from pandas.testing import assert_frame_equal
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()
ds.errs_traceback = False


class TestDsDataClean(BaseDsTest):

	def test_drop_nan(self):
		df1 = pd.DataFrame({"one": ["one","two"], "two":
			["two", None]}, ["1", "2"])
		ds.df = df1
		ds.drop_nan("two")
		df2 = pd.DataFrame({"one": "one", "two": "two"}, ["1"])
		assert_frame_equal(ds.df, df2)
		ds.df = df1
		ds.drop_nan(method="any")
		assert_frame_equal(ds.df, df2)
		ds.df = None
		self.assertErr("TypeError", ds.drop_nan, "one")

	def test_zero_nan(self):
		df1 = pd.DataFrame({"one": ["one","two"], "two":
			["two", 0]}, ["1", "2"])
		ds.df = df1
		ds.zero_nan("two")
		df2 = pd.DataFrame({"one": ["one", "two"], "two":
				["two", nan]}, ["1", "2"])
		assert_frame_equal(ds.df, df2)
		ds.df = df1
		ds.zero_nan("two", "one")
		ds.df = None
		self.assertErr("AttributeError", ds.zero_nan, "one", "two")
		ds.df = df1
		msg = "Column wrong does not exist"
		self.assertWarning(msg, ds.zero_nan, "wrong", "two")
		self.assertErr(None, ds.zero_nan, "wrong", "wrong")

	def test_fill_nan(self):
		df1 = pd.DataFrame({"one": ["one", "two"], "two":
							["two", None]}, ["1", "2"])
		ds.df = df1
		ds.fill_nan("two", "two")
		df2 = pd.DataFrame({"one": ["one", "two"], "two":
							["two", "two"]}, ["1", "2"])
		assert_frame_equal(ds.df, df2)
		ds.df = df1
		ds.fill_nan("two")
		assert_frame_equal(ds.df, df2)
		ds.df = None
		self.assertErr("AttributeError", ds.fill_nan, "val", "two")
		ds.df = df1
		msg = "Can not find column wrong"
		self.assertWarning(msg, ds.fill_nan, "val", "wrong")
		self.assertErr(None, ds.fill_nan, "val", "wrong")

	def test_replace(self):
		df1 = pd.DataFrame({"one": ["one", "two"], "two":
							["two", "two"]}, ["1", "2"])
		ds.df = df1
		ds.replace("one", "two", "three")
		df2 = pd.DataFrame({"one": ["one", "three"], "two":
							["two", "two"]}, ["1", "2"])
		assert_frame_equal(ds.df, df2)
		ds.df = None
		self.assertErr("AttributeError", ds.replace, "one", "two", "three")

	def test_to_int(self):
		df1 = pd.DataFrame({"one": ["one", "two"], "two":
							[1.0, 2.0]}, ["1", "2"])
		ds.df = df1
		ds.to_int("two")
		self.assertEqual(list(ds.df["two"]), [1, 2])
		ds.df = None
		self.assertErr("TypeError", ds.to_int, "two")
		ds.df = pd.DataFrame({"one": ["one", "two"], "two":
							["wrong", 2.0]}, ["1", "2"])
		self.assertErr("ValueError", ds.to_int, "two")

	def test_to_float(self):
		df1 = pd.DataFrame({"one": ["one", "two"], "two":
							[1, 2]}, ["1", "2"])
		ds.df = df1
		ds.to_float("two")
		self.assertEqual(list(ds.df["one"]), [1.0, 2.0])
		ds.df = None
		self.assertErr("AttributeError", ds.to_float, "two")
		ds.df = pd.DataFrame({"one": ["one", "two"], "two": ["wrong", 2]})
		self.assertErr("ValueError", ds.to_float, "two")

	def test_to_type(self):
		df1 = pd.DataFrame({"one": ["one", "two"], "two":[1, 2]})
		ds.df = df1
		ds.to_type(str, "two")
		df2 = pd.DataFrame({"one": ["one", "two"], "two":["1", "2"]})
		df2.astype(str, "two")
		assert_frame_equal(ds.df, df2)
		ds.df = None
		self.assertErr("AttributeError", ds.to_type, str, "two")
		ds.df = pd.DataFrame({"one": ["one", "two"], "two":
							["wrong", 2.0]}, ["1", "2"])
		self.assertErr(None, ds.to_type, "float64", "two")
		self.assertErr(None, ds.to_type, "float64", "wrong")

	def tests_timestamps(self):
		df1 = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2002/12/02"]}, [1, 2])
		ds.df = df1
		ds.timestamps("two",name= "ts")
		df2 = pd.DataFrame({"one": ["one","two"], "ts": [1038700800,
			1038787200], "two":
			["2002/12/01", "2002/12/02"]}, [1, 2])
		ds.df = ds.df.reindex(sorted(ds.df.columns), axis=1)
		assert_frame_equal(ds.df, df2)
		ds.df = df1
		ds.timestamps("two")
		ds.timestamps("two", unit="s", errors="raise")
		ds.df = None
		ds.timestamps("two")
		self.assertRaises(AttributeError)

	def test_date_index(self):
		df1 = pd.DataFrame({"one": ["one","two"], "date":
			["2002/12/01", "2002/12/02"]}, [1, 2])
		ds.df = df1
		ds.dateindex("date")
		index = np.array(['2002-12-01T00:00:00.000000000',
					   '2002-12-02T00:00:00.000000000'],
						dtype="datetime64[ns]")
		np.testing.assert_array_equal(ds.df.index.values, index)
		ds.df = None
		ds.dateindex("date")
		self.assertRaises(TypeError)
		ds.dateindex("wrong")
		self.assertRaises(TypeError)

	def test_index(self):
		df1 = pd.DataFrame([[1, 3], [2, 4]], columns=["one", "two"])
		ds.df = df1
		ds.index("one")
		index = np.array([1, 2])
		np.testing.assert_array_equal(ds.df.index.values, index)
		ds.df = None
		ds.index("one")
		self.assertRaises(TypeError)

	def test_strip(self):
		df = pd.DataFrame({"one": [" 2 ", "1 ", "0 "], 
							"two": [2, 1, 1]})
		ds.df = df
		ds.strip("one")
		df2 = pd.DataFrame({"one": ["2", "1", "0"], 
							"two": [2, 1, 1]})
		assert_frame_equal(ds.df, df2)
		ds.df = None
		ds.strip("one")
		self.assertRaises(TypeError)
		ds.df = "wrong"
		ds.index("one")
		self.assertRaises(AttributeError)

	def test_strip_cols(self):
		df1 = pd.DataFrame([[1, 3], [2, 4]],
						   columns=[" one ", " two "])
		ds.df = df1
		ds.strip_cols()
		df2 = pd.DataFrame([[1, 3], [2, 4]],
						   columns=["one", "two"])
		assert_frame_equal(ds.df, df2)
		ds.df = pd.DataFrame([[1, 3], [2, 4]],
						   columns=["one", None])
		ds.strip_cols()

	def test_roundvals(self):
		df1 = pd.DataFrame([[1.345854, 3.0], [2.1, 4.0]],
						   columns=["one", "two"])
		ds.df = df1
		ds.roundvals("one")
		df2 = pd.DataFrame([[1.35, 3.0], [2.1, 4.0]],
						   columns=["one", "two"])
		assert_frame_equal(ds.df, df2)
		ds.df = df1
		ds.roundvals("wrong")
		self.assertRaises(KeyError)

	def test_format_date(self):
		date = datetime.datetime(2011, 1, 3, 0, 0)
		d2 = ds.format_date_(date)
		self.assertEqual(d2, '2011-01-03 00:00:00')

	def test_fdates(self):
		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		ds.fdate("two")
		df2 = pd.DataFrame({"one": ["one","two"],"two":
			["2002-12-01 00:00:00", "2003-12-02 00:00:00"]}, [1, 2])
		assert_frame_equal(ds.df, df2)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		df2 = pd.DataFrame({"one": ["one","two"],"two":
			["2002", "2003"]}, [1, 2])
		ds.fdate("two", precision="Y")
		assert_frame_equal(ds.df, df2)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		df2 = pd.DataFrame({"one": ["one","two"],"two":
			["2002-12", "2003-12"]}, [1, 2])
		ds.fdate("two", precision="M")
		assert_frame_equal(ds.df, df2)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		df2 = pd.DataFrame({"one": ["one","two"],"two":
			["2002-12-01", "2003-12-02"]}, [1, 2])
		ds.fdate("two", precision="D")
		assert_frame_equal(ds.df, df2)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		df2 = pd.DataFrame({"one": ["one","two"],"two":
			["2002-12-01 00", "2003-12-02 00"]}, [1, 2])
		ds.fdate("two", precision="H")
		assert_frame_equal(ds.df, df2)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		df2 = pd.DataFrame({"one": ["one","two"],"two":
			["2002-12-01 00:00", "2003-12-02 00:00"]}, [1, 2])
		ds.fdate("two", precision="Min")
		assert_frame_equal(ds.df, df2)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		df2 = pd.DataFrame({"one": ["one","two"],"two":
			["2002", "2003"]}, [1, 2])
		ds.fdate("two", format="%Y")
		assert_frame_equal(ds.df, df2)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "2003/12/02"]}, [1, 2])
		ds.fdate("wrong")
		self.assertRaises(KeyError)

		ds.df = pd.DataFrame({"one": ["one","two"], "two":
			["2002/12/01", "wrong"]}, [1, 2])
		ds.fdate("two", "wrong")
		self.assertRaises(TypeError)

		ds.df = None
		ds.fdate("two")
		self.assertRaises(ValueError)

	def test_nulls(self):
		df1 = pd.DataFrame({"one": ["one", "two"], "two":
							["", None]}, ["1", "2"])
		ds.df = df1
		ds.fill_nulls("two")
		df2 = pd.DataFrame({"one": ["one", "two"], "two":
							[nan, nan]}, ["1", "2"])
		assert_frame_equal(ds.df, df2)
		ds.df = None
		ds.fill_nulls("two")
		self.assertRaises(AttributeError)


	def test_nan_empty(self):
		df1 = pd.DataFrame({"one": ["one","two"], "two":
			["two", ""]}, ["1", "2"])
		ds.df = df1
		ds.nan_empty("two")
		df2 = pd.DataFrame({"one": ["one","two"], "two":
			["two", nan]}, ["1", "2"])
		assert_frame_equal(ds.df, df2)
		ds.df = None
		ds.nan_empty("two")
		self.assertRaises(TypeError)




from .test_init import TestDsDataInit
from .test_count import TestDsDataCount
from .test_clean import TestDsDataClean
from .test_search import TestDsDataSearch
from .test_select import TestDsDataSelect
from .test_text import TestDsDataText
from .test_stats import TestDsDataStats
from .test_transform import TestDsDataTransform


class TestDsData(TestDsDataInit, TestDsDataCount, TestDsDataClean,
                 TestDsDataSearch, TestDsDataSelect, TestDsDataText,
                 TestDsDataStats):
    pass
# class TestDsData(TestDsDataTransform):
#    pass





class DbBase():
    """
    A base class for database management
    """

    def _check_db(self) -> bool:
        if self.db is None:
            self.warning("Database not connected")
            return False
        return True



import pandas as pd
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()

ds.errs_traceback = False


class TestDsDataCount(BaseDsTest):

    def test_count(self):
        ds.df = self.df
        msg = "Found 2 rows in the dataframe"
        self.assertOk(msg, ds.count)
        num = ds.count_()
        self.assertEqual(num, 2)
        ds.df = None
        self.assertErr("AttributeError", ds.count)
        self.assertErr("AttributeError", ds.count_)

    def test_count_nulls(self):
        ds.df = pd.DataFrame({"one": None, "two": 2}, ["1", "2"])
        msg = "Found 2 nulls in column one"
        self.assertOk(msg, ds.count_nulls, "one")
        msg = "Can not find column wrong"
        self.assertWarning(msg, ds.count_nulls, "wrong")
        ds.df = None
        self.assertErr("TypeError", ds.count_nulls, "one")

    def test_count_empty(self):
        ds.df = pd.DataFrame({"one": "", "two": 2}, ["1", "2"])
        msg = "Found 2 empty rows in column one"
        self.assertOk(msg, ds.count_empty, "one")
        self.assertErr("KeyError", ds.count_empty, "wrong")
        ds.df = None
        self.assertErr("TypeError", ds.count_empty, "one")

    def test_count_zero(self):
        ds.df = pd.DataFrame({"one": 0, "two": 2}, ["1", "2"])
        msg = "Found 2 zero values in column one"
        self.assertOk(msg, ds.count_zero, "one")
        self.assertErr("KeyError", ds.count_zero, "wrong")
        ds.df = None
        self.assertErr("TypeError", ds.count_zero, "one")

    def test_count_unique(self):
        ds.df = pd.DataFrame({"one": [0, 0, 1], "two": [2, 1, 1]})
        msg = "Found 2 unique values in column one"
        self.assertOk(msg, ds.count_unique_, "one")
        self.assertErr("KeyError", ds.count_unique_, "wrong")



from goerr.colors import colors
from .base import DbBase


class Info(DbBase):
    """
    A class to handle information requests from the database
    """

    def tables(self):
        """Print the existing tables in a database

        :example: ``ds.tables()``
        """
        if self._check_db() == False:
            return
        try:
            pmodels = self._tables()
            if pmodels is None:
                return
            num = len(pmodels)
            s = "s"
            if num == 1:
                s = ""
            msg = "Found " + colors.bold(str(num)) + " table" + s + ":\n"
            msg += "\n".join(pmodels)
            self.info(msg)
        except Exception as e:
            self.err(e, "Can not print tables")

    def tables_(self) -> list:
        """Return a list of the existing tables in a database

        :return: list of the table names
        :rtype: list

        :example: ``tables = ds.tables_()``
        """
        if self._check_db() == False:
            return
        try:
            return self._tables()
        except Exception as e:
            self.err(e, "Can not print tables")

    def _tables(self):
        if self._check_db() == False:
            return
        pmodels = self.db.tables
        if pmodels is None:
            msg = "No tables found in the database"
            self.warning(msg)
            return
        return pmodels

    def table(self, name: str):
        """
        Display info about a table: number of rows
        and columns

        :param name: name of the table
        :type name: str

        :example: ``tables = ds.table("mytable")``
        """
        if self._check_db() == False:
            return
        try:
            res = self.getall(name)
        except Exception as e:
            self.err(e, self.table, "Can not get records from database")
            return
        if res is None:
            self.warning("Table", name, "does not contain any record")
            return
        num = len(res)
        self.info(num, "rows")
        self.info("Fields:", ", ".join(list(res)))



class Resample():
    """
    A class to resample timeseries
    """

    def rsum(self, time_period: str, num_col: str="Number",
             dateindex: str=None):
        """
        Resample and add a sum the main dataframe to a time period

        :param time_period: unit + period: periods are Y, M, D, H, Min, S
        :param time_period: str
        :param num_col: number of the new column, defaults to "Number"
        :param num_col: str, optional
        :param dateindex: column name to use as date index, defaults to None
        :param dateindex: str, optional

        :example: ``ds.rsum("1D")``
        """
        try:
            df = self._resample_("sum", time_period,
                                 num_col, dateindex)
            self.df = df
            if df is None:
                self.err("Can not sum data")
        except Exception as e:
            self.err(e, "Can not sum data")

    def rmean(self, time_period: str, num_col: str="Number",
              dateindex: str=None):
        """
        Resample and add a mean column the main dataframe to a time period

        :param time_period: unit + period: periods are Y, M, D, H, Min, S
        :param time_period: str
        :param num_col: number of the new column, defaults to "Number"
        :param num_col: str, optional
        :param dateindex: column name to use as date index, defaults to None

        :example: ``ds.rmean("1Min")``
        """
        try:
            self.df = self._resample_("mean", time_period,
                                      num_col, dateindex)
        except Exception as e:
            self.err(e, "Can not sum data")

    def _resample_(self, method: str, time_period: str, num_col: str,
                   dateindex: str=None):
        try:
            ds2 = self._duplicate_()
            if dateindex is not None:
                ds2 = self.dateindex_(dateindex)
                if ds2 is None:
                    self.err("Can not process date index")
                    return
            ds2.add(num_col, 1)
            ds2.df = ds2.df.resample(time_period)
            if method == "sum":
                ds2.df = ds2.df.sum()
            elif method == "mean":
                num_vals = ds2.df[num_col].sum()
                ds2.df = ds2.df.mean()
                ds2.df[num_col] = num_vals
            else:
                self.err("Resampling method " + method + " unknown")
                return
            self.ok("Data resampled by", time_period)
            return ds2.df
        except Exception as e:
            self.err(e, "Can not resample data")



# @PydevCodeAnalysisIgnore
import arrow
import numpy as np
import pandas as pd
from numpy.core.numeric import nan
from goerr.colors import colors
from ..errors import Error
from ..messages import Message


class Clean(Error, Message):
    """
    Class to clean the data
    """

    def drop_nan(self, col: str=None, method: str="all", **kwargs):
        """
        Drop rows with NaN values from the main dataframe

        :param col: name of the column, defaults to None. Drops in 
        :type col: str, optional
        :param method: ``how`` param for ``df.dropna``, defaults to "all"
        :type method: str, optional
        :param \*\*kwargs: params for ``df.dropna``
        :type \*\*kwargs: optional

        :example: ``ds.drop_nan("mycol")``
        """
        try:
            if col is None:
                self.df = self.df.dropna(how=method, **kwargs)
            else:
                self.df = self.df[self.df[col].notnull()]
        except Exception as e:
            self.err(e, "Error dropping nan values")

    def nan_empty(self, col: str):
        """
        Fill empty values with NaN values

        :param col: name of the colum
        :type col: str

        :example: ``ds.nan_empty("mycol")``
        """
        try:
            self.df[col] = self.df[col].replace('', nan)
            self.ok("Filled empty values with nan in column " + col)
        except Exception as e:
            self.err(e, "Can not fill empty values with nan")

    def zero_nan(self, *cols):
        """
        Converts zero values to nan values in selected columns

        :param \*cols: names of the colums
        :type \*cols: str, at least one

        :example: ``ds.zero_nan("mycol1", "mycol2")``
        """
        if len(cols) == 0:
            self.warning("Can not nan zero values if a column name "
                         "is not provided")
        df = self._zero_nan(*cols)
        if df is None:
            self.err("Can not fill zero values with nan")
            return
        self.df = df

    def _zero_nan(self, *cols):
        try:
            df = self.df.copy()
            for col in cols:
                if col not in df.columns:
                    self.warning("Column " + col + " does not exist")
                    return
                df = df.replace({col: {0: nan}})
            if len(cols) > 1:
                self.ok("Replaced 0 values by nan in columns",
                        str(cols))
            else:
                self.ok("Replaced 0 values by nan in column",
                        col)
            return df
        except Exception as e:
            self.err(e)

    def fill_nan(self, val: str, *cols):
        """
        Fill NaN values with new values in the main dataframe

        :param val: new value
        :type val: str
        :param \*cols: names of the colums
        :type \*cols: str, at least one

        :example: ``ds.fill_nan("new value", "mycol1", "mycol2")``
        """
        df = self._fill_nan(val, *cols)
        if df is not None:
            self.df = df
        else:
            self.err("Can not fill nan values")

    def _fill_nan(self, val, *cols):
        cols = list(cols)
        if len(cols) == 0:
            cols = self.df.columns.values
        df = None
        try:
            df = self.df.copy()
            for el in cols:
                try:
                    df[el] = df[[el]].fillna(val)
                except KeyError:
                    self.warning("Can not find column", el)
                    return
        except Exception as e:
            self.err(e, "Can not fill nan values")
            return
        self.ok("Filled nan values in columns", ",".join(cols))
        return df

    def fill_nulls(self, col: str):
        """
        Fill all null values with NaN values in a column.
        Null values are ``None`` or en empty string

        :param col: column name
        :type col: str

        :example: ``ds.fill_nulls("mycol")``
        """
        n = [None, ""]
        try:
            self.df[col] = self.df[col].replace(n, nan)
        except Exception as e:
            self.err(e)

    def to_int(self, *cols, **kwargs):
        """
        Convert some column values to integers

        :param \*cols: names of the colums
        :type \*cols: str, at least one
        :param \*\*kwargs: keyword arguments for ``pd.to_numeric``
        :type \*\*kwargs: optional

        :example: ``ds.to_int("mycol1", "mycol2", errors="coerce")``
        """
        try:
            for col in cols:
                self.df[col] = pd.to_numeric(self.df[col], **kwargs)
        except Exception as e:
            self.err(e, "Can not convert column values to integer")
            return
        self.ok("Converted column values to integers")

    def to_float(self, col: str, **kwargs):
        """
        Convert colums values to float

        :param col: name of the colum
        :type col: str, at least one
        :param \*\*kwargs: keyword arguments for ``df.astype``
        :type \*\*kwargs: optional

        :example: ``ds.to_float("mycol1")``
        """
        try:
            self.df[col] = self.df[col].astype(np.float64, **kwargs)
            self.ok("Converted column values to float")
        except Exception as e:
            self.err(e, "Error converting to float")

    def to_type(self, dtype: type, *cols, **kwargs):
        """
        Convert colums values to a given type in the 
        main dataframe

        :param dtype: a type to convert to: ex: ``str``
        :type dtype: type
        :param \*cols: names of the colums
        :type \*cols: str, at least one
        :param \*\*kwargs: keyword arguments for ``df.astype``
        :type \*\*kwargs: optional

        :example: ``ds.to_type(str, "mycol")``
        """
        try:
            allcols = self.df.columns.values
            for col in cols:
                if col not in allcols:
                    self.err("Column " + col + " not found")
                    return
                self.df[col] = self.df[col].astype(dtype, **kwargs)
        except Exception as e:
            self.err(e, "Can not convert to type")

    def fdate(self, *cols, precision: str="S", format: str=None):
        """
        Convert column values to formated date string

        :param \*cols: names of the colums
        :type \*cols: str, at least one
        :param precision: time precision: Y, M, D, H, Min S, defaults to "S"
        :type precision: str, optional
        :param format: python date format, defaults to None
        :type format: str, optional

        :example: ``ds.fdate("mycol1", "mycol2", precision)``
        """

        def formatdate(row):
            return row.strftime(format)

        def convert(row):
            encoded = '%Y-%m-%d %H:%M:%S'
            if precision == "Min":
                encoded = '%Y-%m-%d %H:%M'
            elif precision == "H":
                encoded = '%Y-%m-%d %H'
            elif precision == "D":
                encoded = '%Y-%m-%d'
            elif precision == "M":
                encoded = '%Y-%m'
            elif precision == "Y":
                encoded = '%Y'
            return row.strftime(encoded)

        try:
            for f in cols:
                try:
                    if format is None:
                        self.df[f] = pd.to_datetime(self.df[f]).apply(convert)
                    else:
                        self.df[f] = pd.to_datetime(
                            self.df[f]).apply(formatdate)
                except ValueError as e:
                    self.err(e, "Can not convert date")
                    return
        except KeyError:
            self.warning("Can not find colums " + " ".join(cols))
            return
        except Exception as e:
            self.err(e, "Can not process date col")

    def timestamps(self, col: str, **kwargs):
        """"
        Add a timestamps column from a date column

        :param col: name of the timestamps column to add
        :type col: str
        :param \*\*kwargs: keyword arguments for ``pd.to_datetime``
        :type \*\*kwargs: optional

        :example: ``ds.timestamps("mycol")``
        """
        try:
            name = "Timestamps"
            if "name" in kwargs:
                name = kwargs["name"]
            if "errors" not in kwargs:
                kwargs["errors"] = "coerce"
            if "unit" in kwargs:
                kwargs["unit"] = "ms"
            try:
                self.df[col] = pd.to_datetime(self.df[col], **kwargs)
            except TypeError:
                pass
            ts = []
            for el in self.df[col]:
                ts.append(arrow.get(el).timestamp)
            self.df[name] = ts
        except Exception as e:
            self.err(e, "Can not convert to timestamps")

    def date(self, col: str, **kwargs):
        """
        Convert a column to date type

        :param col: column name
        :type col: str
        :param \*\*kwargs: keyword arguments for ``pd.to_datetime`` 
        :type \*\*kwargs: optional

        :example: ``ds.date("mycol")``
        """
        try:
            self.df[col] = pd.to_datetime(self.df[col], **kwargs)
        except Exception as e:
            self.err(e, "Can not convert to date")

    def dateindex(self, col: str):
        """
        Set a datetime index from a column

        :param col: column name where to index the date from
        :type col: str

        :example: ``ds.dateindex("mycol")``
        """
        df = self._dateindex(col)
        if df is None:
            self.err("Can not create date index")
            return
        self.df = df
        self.ok("Added a datetime index from column", col)

    def _dateindex(self, col: str) -> pd.DataFrame:
        try:
            index = pd.DatetimeIndex(self.df[col])
            df = self.df.set_index(index)
            return df
        except Exception as e:
            self.err(e, "Can not index with column " +
                     colors.bold(col))

    def index(self, col: str):
        """
        Set an index to the main dataframe

        :param col: column name where to index from
        :type col: str

        :example: ``ds.index("mycol")``
        """
        df = self._index(col)
        if df is None:
            self.err("Can not create index")
            return
        self.df = df

    def _index(self, col: str) -> pd.DataFrame:
        try:
            df = self.df.set_index(self.df[col])
        except Exception as e:
            self.err(e, self._index, "Can not find column " +
                     colors.bold(col) + " in data")
            return
        self.ok("Added an index from column", col)
        return df

    def strip(self, col: str):
        """
        Remove leading and trailing white spaces in a column's values

        :param col: name of the column
        :type col: str

        :example: ``ds.strip("mycol")``
        """
        def remove_ws(row):
            val = str(row[col])
            if " " in val.startswith(" "):
                row[col] = val.strip()
            return row

        try:
            self.df.apply(remove_ws)
        except Exception as e:
            self.err(e, "Can not remove white space in column")
            return
        self.ok("White space removed in column values")

    def strip_cols(self):
        """
        Remove leading and trailing white spaces in columns names

        :example: ``ds.strip_cols()``
        """
        cols = {}
        skipped = []
        for col in self.df.columns.values:
            try:
                cols[col] = col.strip()
            except Exception:
                skipped.append(str(col))
        self.df = self.df.rename(columns=cols)
        self.ok("White spaces removed in columns names")
        if len(skipped) > 0:
            self.info("Skipped columns", ','.join(
                skipped), "while removing white spaces")

    def roundvals(self, col: str, precision: int=2):
        """
        Round floats in a column. Numbers are going to be
        converted to floats if they are not already

        :param col: column name
        :type col: str
        :param precision: float precision, defaults to 2
        :param precision: int, optional

        :example: ``ds.roundvals("mycol")``
        """
        try:
            self.df[col] = self.df[col].astype("float64")
            self.df[col] = self.df[col].apply(lambda x: round(x, precision))
        except Exception as e:
            self.err(e, "Can not round column values")
            return
        self.ok("Rounded values in column " + col)

    def replace(self, col: str, searchval: str, replaceval: str):
        """
        Replace a value in a column in the main dataframe

        :param col: column name
        :type col: str
        :param searchval: value to replace
        :type searchval: str
        :param replaceval: new value
        :type replaceval: str

        :example: ``ds.replace("mycol", "value", "new_value")``
        """
        try:
            self.df[col] = self.df[col].replace(searchval, replaceval)
        except Exception as e:
            self.err(e, "Can not replace value in column")

    def creplace(self, col: str, searchval: str, replaceval: str):
        """
        Replace a value in a column in the main dataframe
        if the value is contained

        :param col: column name
        :type col: str
        :param searchval: value to replace
        :type searchval: str
        :param replaceval: new value
        :type replaceval: str

        :example: ``ds.creplace("mycol", "value", "new_value")``
        """
        try:
            self.df[col] = self.df[col].str.replace(searchval,
                                                    replaceval)
        except Exception as e:
            self.err("Can not replace value in column")

    def format_date_(self, date):
        """
        Formats a date
        """
        return date.strftime('%Y-%m-%d %H:%M:%S')



import folium
from folium.features import DivIcon


class Map():

    def __init__(self):
        """
        Initialize
        """
        self.dsmap = None

    def tmarker(self, lat, long, text, color=None, icon=None, style=None):
        """
        Returns the map with a text marker to the default map
        """
        try:
            self.dsmap = self._marker(
                lat, long, text, self.dsmap, color, icon, True, style)
            return self.dsmap
        except Exception as e:
            self.err(e, self.tmarker, "Can not get text marker")

    def tmarker_(self, lat, long, text, pmap,
                 color=None, icon=None, style=None):
        """
        Returns the map with a text marker to the default map
        """
        try:
            xmap = self._marker(
                lat, long, text, pmap, color, icon, True, style)
            return xmap
        except Exception as e:
            self.err(e, self.tmarker_, "Can not get text marker")

    def marker_(self, lat, long, text, pmap, color=None, icon=None):
        """
        Returns the map with a marker to the default map
        """
        try:
            xmap = self._marker(lat, long, text, pmap, color, icon)
            return xmap
        except Exception as e:
            self.err(e, self.marker_, "Can not get marker")

    def marker(self, lat, long, text, color=None, icon=None):
        """
        Set the main map with a marker to the default map
        """
        try:
            self.dsmap = self._marker(lat, long, text, self.dsmap, color, icon)
        except Exception as e:
            self.err(e, self.marker, "Can not get marker")

    def _marker(self, lat, long, text, xmap, color=None, icon=None,
                text_mark=False, style=None):
        """
        Adds a marker to the default map
        """
        kwargs = {}
        if icon is not None:
            kwargs["icon"] = icon
        if color is not None:
            kwargs["color"] = color
        if style is None:
            style = "font-size:18pt;font-weight:bold;" + \
                "color:black;border-radius:0.5"
        try:
            xicon1 = folium.Icon(**kwargs)
            if text_mark is True:
                xicon = DivIcon(
                    icon_size=(150, 36),
                    icon_anchor=(0, 0),
                    html='<div style="' + style + '">' + text + '</div>',
                )
                folium.Marker([lat, long], popup=text,
                              icon=xicon).add_to(xmap)
            folium.Marker([lat, long], popup=text,
                          icon=xicon1).add_to(xmap)
            return xmap
        except Exception as e:
            self.err(e, self._marker, "Can not get marker")

    def map_(self, lat, long, zoom=13, tiles="map"):
        """
        Returns a map
        """
        try:
            return self._map(lat, long, zoom, tiles)
        except Exception as e:
            self.err(e, self.map_, "Can not get map")

    def amap(self, lat, long, zoom=13, tiles="map"):
        """
        Sets a map
        """
        try:
            self.dsmap = self._map(lat, long, zoom, tiles)
        except Exception as e:
            self.err(e, self.map, "Can not get map")

    def _map(self, lat, long, zoom, tiles):
        """
        Returns a map
        """
        if tiles == "map":
            tiles = "OpenStreetMap"
        elif tiles == "terrain":
            tiles = "Stamen Terrain"
        elif tiles == "basic":
            tiles = "Stamen Toner"
        try:
            xmap = folium.Map(
                location=[lat, long],
                tiles=tiles,
                zoom_start=zoom
            )
            return xmap
        except Exception as e:
            self.err(e, self._map, "Can not make map")



# @PydevCodeAnalysisIgnore
from numpy.core.numeric import nan
import dataset
from stuf import stuf
import pandas as pd
from .infos import Info
from .relations import Relation
from .insert import Insert
from .select import Select
from .influxdb import InfluxDb


class Db(Info, Insert, Relation, Select, InfluxDb):
    """
    Class for manipulating databases
    """

    def __init__(self):
        """
        Initialize with an empty database
        """
        self.db = db

    def connect(self, url: str):
        """Connect to the database and set it as main database

        :param url: path to the database, uses the Sqlalchemy format
        :type url: str

        :example: ``ds.connect("sqlite:///mydb.slqite")``
        """
        try:
            self.db = dataset.connect(url, row_type=stuf)
        except Exception as e:
            self.err(e, "Can not connect to database")
            return
        if self.db is None:
            self.err("Database " + url + " not found")
            return
        self.ok("Db", self.db.url, "connected")

    def load(self, table: str):
        """Set the main dataframe from a table's data

        :param table: table name
        :type table: str

        :example: ``ds.load("mytable")``
        """
        if self._check_db() is False:
            return
        if table not in self.db.tables:
            self.warning("The table " + table + " does not exists")
            return
        try:
            self.start("Loading data from table " + table)
            res = self.db[table].all()
            self.df = pd.DataFrame(list(res))
            self.end("Data loaded from table " + table)
        except Exception as e:
            self.err(e, "Can not load table " + table)

    def load_django(self, query: "django query"):
        """Load the main dataframe from a django orm query

        :param query: django query from a model
        :type query: django query

        :example: ``ds.load_django(Mymodel.objects.all())``
        """
        try:
            self.df = self._load_django(query)
        except Exception as e:
            self.err(e, "Can not load data from query")

    def load_django_(self, query: "django query") -> "Ds":
        """Returns a DataSwim instance from a django orm query

        :param query: django query from a model
        :type query: django query
        :return: a dataswim instance with data from a django query
        :rtype: Ds

        :example: ``ds2 = ds.load_django_(Mymodel.objects.all())``
        """
        try:
            df = self._load_django(query)
            return self.clone_(df)
        except Exception as e:
            self.err(e, "Can not load data from query")

    def _load_django(self, query: "django query"):
        try:
            df = pd.DataFrame(list(query.values()))
            return df
        except Exception as e:
            self.err(e)



# -*- coding: utf-8 -*-
import pandas as pd
from pandas.testing import assert_series_equal
from sklearn import linear_model
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()

ds.errs_traceback = False


class TestDsDataStats(BaseDsTest):

    def test_lreg(self):
        df1 = pd.DataFrame([["2001/01/01", 1], ["2001/01/02", 2],
                            ["2001/01/03", 3], ["2001/01/04", 3]],
                           columns=["date", "value"])
        ds.df = df1
        ds.timestamps("date")
        ds.lreg("Timestamps", "value")
        expected = pd.Series([1.2, 1.9, 2.6, 3.3], name="Regression")
        assert_series_equal(ds.df["Regression"], expected)
        ds.df = None
        self.assertErr("TypeError", ds.lreg, "x", "y")

    def test_cvar(self):
        ds.df = pd.DataFrame([1, 3, 3, 1])
        cv = ds.cvar_(0)
        self.assertEqual(cv, 50.0)
        ds.df = None
        self.assertErr("TypeError", ds.cvar_, 0)



import os
import unittest
from io import StringIO
from contextlib import redirect_stdout
import pandas as pd
from goerr.colors import colors
from goerr.testing import assert_err
from dataswim import Ds


ds = Ds()


class BaseDsTest(unittest.TestCase):

    def setUp(self):
        self.df = pd.DataFrame({"one": 1, "two": 2}, ["1", "2"])
        self.path = os.path.dirname(os.path.realpath(__file__))

    def debug(self, *elements):
        print("-------------------------------------------")
        ds.debug(*elements)
        print("-------------------------------------------")

    def assertErr(self, err_type, func, *args, **kwargs):
        assert_err(err_type, func, *args, **kwargs)

    def assertWarning(self, txt, func, *args, **kwargs):
        self._assertPrintMsg(True, "warning", txt, func, *args, **kwargs)

    def assertOk(self, txt, func, *args, **kwargs):
        self._assertPrintMsg(False, "ok", txt, func, *args, **kwargs)

    def _assertPrint(self, contains, expected, func, *args, **kwargs):
        f = StringIO()
        with redirect_stdout(f):
            func(*args, **kwargs)
        displayed = f.getvalue()
        if contains is True:
            self.assertIn(expected, displayed)
        else:
            self.assertEqual(displayed, expected)

    def _assertPrintMsg(self, contains, msg_class, txt, func, *args, **kwargs):
        if msg_class == "ok":
            label = colors.green("OK")
        elif msg_class == "warning":
            label = colors.yellow("WARNING")
        elif msg_class == "info":
            label = colors.blue("INFO")
        elif msg_class == "progress":
            label = colors.purple("Progress")
        elif msg_class == "start":
            label = colors.purple("START")
        elif msg_class == "end":
            label = colors.purple("END")
        expected = ds.msg_(label, txt)+"\n"
        self._assertPrint(contains, expected, func, *args, **kwargs)



# @PydevCodeAnalysisIgnore
import pandas as pd
import deepdish as dd
from .views import View
from .clean import Clean
from .count import Count
from .select import Select
from .transform import Transform
from .export import Export
from .search import Search
from .stats import Stats
from .text import Text
from .copy import Copy
from ..errors import Error
from ..messages import Message


class Df(Select, View, Transform, Clean, Count,
         Export, Search, Stats, Text, Copy):
    """
    Class for manipulating dataframes
    """

    def backup(self):
        """
        Backup the main dataframe
        """
        try:
            self.backup_df = self.df.copy()
        except Exception as e:
            self.err(e, "Can not backup data")
            return
        self.ok("Dataframe backed up")

    def restore(self):
        """
        Restore the main dataframe
        """
        if self.backup_df is None:
            self.warning("No dataframe is backed up: nothing restore")
            return
        self.df = self.backup_df
        self.ok("Dataframe is restored")

    def load_json(self, path, **kwargs):
        """Load data in the main dataframe from json

        :param filepath: url of the csv file to load,
                                         can be absolute if it starts with ``/``
                                         or relative if it starts with ``./``
        :type filepath: str

        :param kwargs: keyword arguments to pass to
                                                   Pandas ``read_json`` function

        :example: ``ds.load_json("./myfile.json")``
        """
        try:
            df = pd.read_json(path, **kwargs)
            self.df = df
        except Exception as e:
            self.err(e, "Can not load json")

    def load_h5(self, filepath):
        """Load a Hdf5 file to the main dataframe

        :param filepath: url of the csv file to load,
                                         can be absolute if it starts with ``/``
                                         or relative if it starts with ``./``
        :type filepath: str

        :example: ``ds.load_h5("./myfile.hdf5")``
        """
        try:
            self.start("Loading Hdf5 data...")
            self.df = dd.io.load(filepath)
            self.end("Finished loading Hdf5 data")
        except Exception as e:
            self.err(e, "Can not load Hdf5 file")

    def load_excel(self, filepath, **kwargs):
        """Set the main dataframe with the content of an Excel file

        :param filepath: url of the csv file to load,
                        can be absolute if it starts with ``/``
                        or relative if it starts with ``./``
        :type filepath: str

        :param kwargs: keyword arguments to pass to 
                            Pandas ``read_excel`` function

        :example: ``ds.load_excel("./myfile.xlsx")``
        """
        try:
            df = pd.read_excel(filepath, **kwargs)
            if len(df.index) == 0:
                self.warning("Empty Excel file. Can not set the dataframe.")
                return
            self.df = df
        except Exception as e:
            self.err(e, "Can not load Excel file")

    def load_csv(self, url, **kwargs):
        """Loads csv data in the main dataframe

        :param url: url of the csv file to load:
                                can be absolute if it starts with ``/``
                                or relative if it starts with ``./``
        :type url: str
        :param kwargs: keyword arguments to pass to Pandas
                                    ``read_csv`` function

        :example: ``ds.load_csv("./myfile.csv")``
        """
        self.start("Loading csv...")
        try:
            if self.datapath is not None and url.startswith("/") is False:
                url = self.datapath + "/" + url
            df = pd.read_csv(url, **kwargs)
            self.df = df
        except FileNotFoundError:
            msg = "File " + url + " not found"
            self.warning(msg)
            return
        except Exception as e:
            self.err(e, "Can not load csv file")
            return
        self.end("Finished loading csv")

    def dateparser(self, dformat='%d/%m/%Y'):
        """
        Returns a date parser for pandas
        """

        def dateparse(dates):
            return [pd.datetime.strptime(d, dformat) for d in dates]

        return dateparse



# -*- coding: utf-8 -*-
import pandas as pd
from pandas.testing import assert_frame_equal
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()

ds.errs_traceback = False


class TestDsDataTransform(BaseDsTest):

    def test_keep(self):
        df1 = pd.DataFrame([[1,2], [1,2]], columns = ["one", "two"])
        ds.df = df1
        ds.keep("one")
        df2 = pd.DataFrame([1,1], columns = ["one"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.keep_("one")
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        self.assertErr("TypeError", ds.keep, "one")
        self.assertErr("TypeError", ds.keep_, "one")

    def test_rsum(self):
        df1 = pd.DataFrame([["2001/01/01", 1], ["2001/02/01", 5],
                            ["2001/02/15", 3], ["2001/02/28", 3]],
                            columns = ["date", "value"])
        ds.df = df1
        ds.rsum(time_period="1M", dateindex="date", index_col=None)
        index = pd.DatetimeIndex(["2001/01/31", "2001/02/28"], name="date")
        df2 = pd.DataFrame([[1, 1], [11, 3]], columns = ["value", "Number"],
                           index=index)
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.rsum_(time_period="1M", dateindex="date", index_col=None)
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        self.assertErr(None, ds.rsum, time_period="1M",
                       dateindex="date")
        self.assertErr(None, ds.rsum_, time_period="1M",
                       dateindex="date")

    def test_rmean(self):
        df1 = pd.DataFrame([["2001/01/01", 1], ["2001/01/02", 2],
                            ["2001/02/15", 1], ["2001/02/28", 2]],
                            columns = ["date", "value"])
        ds.df = df1
        ds.rmean(time_period="1M", dateindex="date", index_col=None)
        index = pd.DatetimeIndex(["2001/01/31", "2001/02/28"], name="date")
        df2 = pd.DataFrame(pd.DataFrame([[1.5, 2], [1.5, 2]],
                  columns = ["value", "Number"], index=index))
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.rmean_(time_period="1M", dateindex="date", index_col=None)
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        self.assertErr(None, ds.rmean, time_period="1M",
                       dateindex="date")
        self.assertErr(None, ds.rmean_, time_period="1M",
                       dateindex="date")

    def test_resample(self):
        df1 = pd.DataFrame([["2001/01/01", 1], ["2001/01/02", 2],
                            ["2001/02/15", 1], ["2001/02/28", 2]],
                            columns = ["date", "value"])
        ds.df = df1
        ds.dateindex("date")
        ds.df = ds._resample_(time_period="1M", method="sum", num_col=None,
                             fill_col=None, dateindex=None, index_col=None)
        index = pd.DatetimeIndex(["2001/01/31", "2001/02/28"], name="date")
        df2 = pd.DataFrame([[3], [3]], columns = ["value"], index=index)
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds.dateindex("date")
        ds.df = ds._resample_(time_period="1M", method="mean", num_col="num",
                             fill_col=None, dateindex=None, index_col=None)
        ds.df = df1
        ds.dateindex("date")
        self.assertErr(None, ds._resample_, time_period="1M",
                       method="sum", num_col=None, fill_col=None,
                       dateindex="wrong", index_col=None)
        self.assertErr(None, ds._resample_, time_period="1M",
                       method="wrong", num_col=None, fill_col=None,
                       dateindex="date", index_col=None)
        self.assertErr(None, ds._resample_, time_period="1M",
                       method="sum", num_col="n", fill_col="wrong",
                       dateindex="wrong", index_col=None)

    def test_sum(self):
        df1 = pd.DataFrame([1, 2])
        ds.df = df1
        self.assertEqual(ds.sum_(0), 3)
        ds.df = None
        self.assertErr("TypeError", ds.sum_, 0)

    def test_reverse(self):
        df1 = pd.DataFrame([1, 2], index=[1,2])
        ds.df = df1
        df2 = pd.DataFrame([2, 1], index=[2, 1])
        ds.reverse()
        assert_frame_equal(ds.df, df2)
        ds.df = None
        self.assertErr("AttributeError", ds.reverse)

    def test_sort(self):
        df1 = pd.DataFrame([2, 1])
        ds.df = df1
        df2 = pd.DataFrame([1, 2])
        ds.sort(0)
        self.assertEqual(list(ds.df), list(df2))
        ds.df = None
        self.assertErr("AttributeError", ds.sort, 0)

    def test_apply(self):
        df1 = pd.DataFrame([[1,2], [1,2]], columns = ["one", "two"])
        ds.df = df1

        def ap(row):
            return row + 1

        ds.apply(ap, "one")
        df2 = pd.DataFrame([[2,2], [2,2]], columns = ["one", "two"])
        assert_frame_equal(ds.df, df2)
        ds.df = pd.DataFrame([[1,2], [1,2]], columns = ["one", "two"])
        ds.apply(ap)
        df3 = pd.DataFrame([[2,3], [2,3]], columns = ["one", "two"])
        assert_frame_equal(ds.df, df3)
        ds.df = None
        self.assertErr("TypeError", ds.apply, ap, "one")

    def test_replace(self):
        df1 = pd.DataFrame([[1,2], [1, 2]], columns=["one", "two"])
        ds.df = df1
        df2 = pd.DataFrame([[2,2], [2, 2]], columns=["one", "two"])
        ds.replace("one", 1, 2)
        self.assertEqual(list(ds.df), list(df2))
        ds.df = None
        self.assertErr("TypeError", ds.replace, "one", 1, 2)

    def test_pivot(self):
        df1 = pd.DataFrame([["n2", 1], ["n1", 2], ["n1", 1]],
                     columns = ["name", "val"])
        ds.df = df1
        ds.pivot("name")
        df2 = pd.DataFrame([1.5, 1.0], columns = ["val"], index=["n1", "n2"])
        assert_frame_equal(ds.df, df2, check_names=False)
        ds.df = df1
        ds2 = ds.pivot_("name")
        assert_frame_equal(ds2.df, df2, check_names=False)
        ds.df = None
        self.assertErr(None, ds.pivot_, "name")
        self.assertErr(None, ds.pivot, "name")

    def test_concat(self):
        df1 = pd.DataFrame([1,2])
        df2 = pd.DataFrame([3,4])
        ds.df = df1
        ds2 = ds.new_(df2)
        ds.concat(ds, ds2)
        df3 = pd.DataFrame([1,2,3,4])
        self.assertEqual(list(ds.df), list(df3))
        ds.df = df1
        ds2 = ds.concat_(ds, ds2)
        self.assertEqual(list(ds2.df), list(df3))
        ds2 = None
        self.assertErr(None, ds.concat_, ds, ds2)
        self.assertErr(None, ds.concat, ds, ds2)

    def test_split(self):
        df1 = pd.DataFrame([[1], [2], [1], [2]], columns=["val"])
        ds.df = df1
        dsd = ds.split_("val")
        d = "{1: <DataSwim object | 2 rows>, 2: <DataSwim object | 2 rows>}"
        self.assertEqual(str(dsd), d)


    def test_drop(self):
        df1 = pd.DataFrame([[1,2], [1,2]], columns = ["one", "two"])
        ds.df = df1
        ds.drop("two")
        df2 = pd.DataFrame([1,1], columns = ["one"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        msg = "Column wrong not found. Aborting"
        self.assertWarning(msg, ds.drop, "wrong")
        ds.df = None
        self.assertErr("AttributeError", ds.drop, "wrong")

    def test_exclude(self):
        df1 = pd.DataFrame([[2,2], [1,2]], columns = ["one", "two"])
        ds.df = df1
        ds.exclude("one", 1)
        df2 = pd.DataFrame([[2, 2]], columns = ["one", "two"])
        assert_frame_equal(ds.df, df2)
        self.assertErr("KeyError", ds.exclude, "wrong", 0)



import numpy
import pytablewriter
import deepdish as dd


class Export():
    """
    Class to handle data exportation
    """

    def to_html_(self) -> str:
        """Convert the main dataframe to html

        :return: html data
        :rtype: str

        :example: ``ds.to_html_()``
        """
        try:
            renderer = pytablewriter.HtmlTableWriter
            data = self._build_export(renderer)
            return data
        except Exception as e:
            self.err(e, "Can not convert data to html")

    def to_json_(self) -> str:
        """Convert the main dataframe to json

        :return: json data
        :rtype: str

        :example: ``ds.to_json_()``
        """
        try:
            renderer = pytablewriter.JsonTableWriter
            data = self._build_export(renderer)
            return data
        except Exception as e:
            self.err(e, "Can not convert data to json")

    def to_javascript_(self, table_name: str="data") -> str:
        """Convert the main dataframe to javascript code

        :param table_name: javascript variable name, defaults to "data"
        :param table_name: str, optional
        :return: a javascript constant with the data
        :rtype: str

        :example: ``ds.to_javastript_("myconst")``
        """
        try:
            renderer = pytablewriter.JavaScriptTableWriter
            data = self._build_export(renderer, table_name)
            return data
        except Exception as e:
            self.err(e, "Can not convert data to javascript code")

    def to_markdown_(self) -> str:
        """Convert the main dataframe to markdown

        :return: markdown data
        :rtype: str

        :example: ``ds.to_markdown_()``
        """
        try:
            renderer = pytablewriter.MarkdownTableWriter
            data = self._build_export(renderer)
            return data
        except Exception as e:
            self.err(e, "Can not convert data to markdown")

    def to_rst_(self) -> str:
        """Convert the main dataframe to restructured text

        :return: rst data
        :rtype: str

        :example: ``ds.to_rst_()``
        """
        try:
            renderer = pytablewriter.RstGridTableWriter
            data = self._build_export(renderer)
            return data
        except Exception as e:
            self.err(e, "Can not convert data to restructured text")

    def to_python_(self, table_name: str="data") -> list:
        """Convert the main dataframe to python a python list

        :param table_name: python variable name, defaults to "data"
        :param table_name: str, optional
        :return: a python list of lists with the data
        :rtype: str

        :example: ``ds.to_python_("myvar")``
        """
        try:
            renderer = pytablewriter.PythonCodeTableWriter
            data = self._build_export(renderer, table_name)
            return data
        except Exception as e:
            self.err(e, "Can not convert data to python list")

    def to_numpy_(self, table_name: str="data") -> numpy.array:
        """Convert the main dataframe to a numpy array

        :param table_name: name of the python variable, defaults to "data"
        :param table_name: str, optional
        :return: a numpy array
        :rtype: np.array

        :example: ``ds.to_numpy_("myvar")``
        """
        try:
            renderer = pytablewriter.NumpyTableWriter
            data = self._build_export(renderer, table_name)
            return data
        except Exception as e:
            self.err(e, "Can not convert data to numpy array")

    def to_records_(self) -> dict:
        """Returns a list of dictionary records from the main dataframe

        :return: a python dictionnary with the data
        :rtype: str

        :example: ``ds.to_records_()``
        """
        try:
            dic = self.df.to_dict(orient="records")
            return dic
        except Exception as e:
            self.err(e, "Can not convert data to records")

    def to_csv(self, filepath: str, index: bool=False, **kwargs):
        """Write the main dataframe to a csv file

        :param filepath: path of the file to save
        :type filepath: str
        :param index: [description], defaults to False
        :param index: bool, optional
        :param \*args: arguments to pass to ``pd.to_csv``

        :example: ``ds.to_csv_("myfile.csv", header=false)``
        """
        try:
            self.start("Saving data to "+filepath + " ...")
            if self.datapath is not None:
                if filepath.startswith("/") is False and \
                        filepath.startswith(".") is False:
                    filepath = self.datapath + "/" + filepath
            self.df.to_csv(filepath, encoding='utf-8', index=index, **kwargs)
            self.end("Data exported to", filepath)
        except Exception as e:
            self.err(e, "Can not convert data to csv")

    def to_excel(self, filepath: str, title: str):
        """Write the main dataframe to an Excell file

        :param filepath: path of the Excel file to write
        :type filepath: str
        :param title: Title of the stylesheet
        :type title: str

        :example: ``ds.to_excel_("./myfile.xlsx", "My data")``
        """
        try:
            self.start("Saving data to Excell file: "+filepath + " ...")
            writer = pytablewriter.ExcelXlsxTableWriter()
            writer.from_dataframe(self.df)
            writer.open(filepath)
            writer.make_worksheet(title)
            writer.write_table()
            writer.close()
            self.end("File exported to", filepath)
        except Exception as e:
            self.err(e, "Can not convert data to Excel")

    def to_hdf5(self, filepath: str):
        """Write the main dataframe to Hdf5 file

        :param filepath: path where to save the file
        :type filepath: str

        :example: ``ds.to_hdf5_("./myfile.hdf5")``
        """
        try:
            self.start("Saving data to Hdf5...")
            dd.io.save(filepath, self.df)
            self.end("Finished saving Hdf5 data")
        except Exception as e:
            self.err(e, "Can not convert data to Hdf5")

    def _build_export(self, renderer, table_name=None):
        writer = renderer()
        writer.from_dataframe(self.df)
        if table_name is not None:
            writer.table_name = table_name
        data = writer.write_table()
        return data



# -*- coding: utf-8 -*-
import pandas as pd
from pandas.testing import assert_frame_equal
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()

ds.errs_traceback = False


class TestDsDataText(BaseDsTest):

    def test_flat(self):
        df1 = pd.DataFrame([["one", "one"], ["two", "two"]],
                           columns=["one", "two"])
        ds.df = df1
        data = ds.flat_("one")
        self.assertEqual(data, '0 one 1 two')
        data = ds.flat_("one", False)
        self.assertEqual(data, 'one two')
        ds.df = None
        data = ds.flat_("one")
        self.assertRaises(TypeError)

    def test_mfw(self):
        df1 = pd.DataFrame([["one", "one"], ["two", "two"], ["one", "three"]],
                           columns=["one", "two"])
        ds.df = df1
        ds2 = ds.mfw_("one")
        df2 = pd.DataFrame([[2, "one"], [1, "two"]],
                           columns=["Frequency", "Word"]).set_index('Word')
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        _ = ds.mfw_("one")
        self.assertRaises(TypeError)



from ..errors import Error
from ..messages import Message


class Copy(Error, Message):
	
    def _duplicate_(self, df=None, db=None, quiet=True):
        try:
            if db is None:
                db = self.db
            if df is None:
                if self.df is None:
                    self.err("The main dataframe is empty and no dataframe"
                             " was provided: please provide a dataframe as argument")
                    return
                df = self.df.copy()
            ds2 = self.new_(df, db, quiet=True, nbload_libs=False)
            ds2.db = self.db
            ds2.x = self.x
            ds2.y = self.y
            ds2.chart_obj = self.chart_obj
            ds2.chart_opts = self.chart_opts
            ds2.chart_style = self.chart_style
            ds2.label = self.label
            ds2.reports = self.reports
            ds2.report_engines = self.report_engines
            ds2.backup_df = self.backup_df
            ds2.autoprint = self.autoprint
            ds2.errors_handling = self.errors_handling
            ds2.datapath = self.datapath
            ds2.report_path = self.report_path
            ds2.static_path = self.static_path
            ds2.quiet = self.quiet
        except Exception as e:
            self.err(e, "Can not duplicate instance")
            return
        if quiet is False:
            self.ok("A duplicated instance was created")
        return ds2

    def clone_(self, quiet=False):
        """
        Clone the DataSwim instance
        """
        ds2 = self._duplicate_(quiet=True)
        if ds2 is None:
            self.err("Can not clone instance")
        else:
            if quiet is False:
                self.ok("Instance cloned")
        return ds2



# -*- coding: utf-8 -*-
import pandas as pd
from pandas.testing import assert_frame_equal
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()

ds.errs_traceback = False


class TestDsDataSearch(BaseDsTest):

    def test_contains(self):
        df1 = pd.DataFrame([["one x", "three"], ["two", "four"]])
        ds.df = df1
        ds.contains(0, "one")
        df2 = pd.DataFrame([["one x", "three"]])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.contains_(0, "one")
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        res = ds.contains_(0, "one")
        self.assertRaises(TypeError)
        ds.contains(0, "one")
        self.assertRaises(TypeError)

    def test_exact(self):
        df1 = pd.DataFrame([["one", "three"], ["two", "four"]])
        ds.df = df1
        ds.exact(0, "one")
        df2 = pd.DataFrame([["one", "three"]])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.exact_(0, "one")
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        res = ds.exact_(0, "one")
        self.assertRaises(TypeError)
        ds.exact(0, "one")
        self.assertRaises(TypeError)



from .base import DbBase


class Insert(DbBase):
    """
    A class to handle data ingestion by the database
    """

    def __init__(self, db=None):
        """
        Initialize with an empty db
        """
        self.db = db

    def insert(self, table, records, create_cols=False, dtypes=None):
        """
        Insert one or many records in the database from a dictionary or a list of dictionaries
        """
        if self._check_db() is False:
            return
        try:
            table = self.db[table]
        except Exception as e:
            self.err(e, "Can not find table " + table)
        t = type(records)
        if t == dict:
            func = table.insert
        elif t == list:
            func = table.insert_many
        else:
            msg = "Rows datatype " + \
                str(t) + " not valid: use a list or a dictionary"
            self.err(msg)
        if create_cols is True:
            try:
                func(records, ensure=True, types=dtypes)
            except Exception as e:
                self.err(e, "Can not insert create columns and insert data")
            return
        else:
            try:
                func(records, types=dtypes)
            except Exception as e:
                self.err(e, "Can not insert data")
            return
        self.ok("Rows inserted in the database")

    def upsert(self, table: str, record: dict, create_cols: bool=False,
               dtypes: list=None, pks=["id"], namefields=["id"]):
        """
        Upsert a record in a table
        """
        try:
            self.db[table].upsert(record, pks, create_cols, dtypes)
        except Exception as e:
            self.err(e, "Can not upsert data")
            return
        names = ""
        for el in namefields:
            names += " " + record[el]
        self.ok("Upserted record"+names)

    def update_table(self, table, pks=['id']):
        """
        Update records in a database table from the main dataframe
        """
        if self._check_db() is False:
            return
        try:
            table = self.db[table]
        except Exception as e:
            self.err(e, "Can not find table " + table)
        if self.db is None:
            msg = "Please connect a database before or provide a database url"
            self.err(msg)
            return
        recs = self.to_records_()
        for rec in recs:
            table.insert_ignore(rec, pks, ensure=True)
        self.ok("Data updated in table", table)

    def to_db(self, table, dtypes=None):
        """
        Save the main dataframe to the database
        """
        if self._check_db() is False:
            return
        if table not in self.db.tables:
            self.db.create_table(table)
            self.info("Created table ", table)
        self.start("Saving data to database table " + table + " ...")
        recs = self.to_records_()
        self.insert(table, recs, dtypes)
        self.end("Data inserted in table", table)



import pandas as pd
from numpy import nan


class Calculations():
    """
    Class to make calculations on the data
    """

    def diffn(self, diffcol: str, name: str="Diff"):
        """
        Add a diff column to the main dataframe: calculate the diff
        from the next value

        :param diffcol: column to diff from
        :type diffcol: str
        :param name: diff column name, defaults to "Diff"
        :type name: str, optional

        :example: ``ds.diffn("Col 1", "New col")``
        """
        try:
            vals = []
            i = 0
            for _, row in self.df.iterrows():
                current = row[diffcol]
                try:
                    nextv = self.df[diffcol].iloc[i + 1]
                except Exception:
                    vals.append(nan)
                    continue
                val = nextv - current
                vals.append(round(val, 1))
                i += 1
            self.add("Diff", vals)
        except Exception as e:
            self.err(e, self._append, "Can not diff column")
            return
        self.ok("Diff column " + name + " added to the dataframe")

    def diffp(self, diffcol: str, name: str="Diff"):
        """
        Add a diff column to the main dataframe: calculate the diff
        from the previous value

        :param diffcol: column to diff from
        :type diffcol: str
        :param name: diff column name, defaults to "Diff"
        :type name: str, optional

        :example: ``ds.diffp("Col 1", "New col")``
        """
        try:
            df = self.df.copy()
            previous = 0
            i = 0
            vals = [df[diffcol].iloc[0]]
            for _, row in df.iterrows():
                val = row[diffcol] - previous
                new = round(val, 1)
                previous = row[diffcol]
                if i == 0:
                    vals = [0]
                else:
                    vals.append(new)
                i = 1
            self.df = df
            self.add(name, vals)
        except Exception as e:
            self.err(e, self._append, "Can not diff column")
            return
        self.ok("Diff column " + name + " added to the dataframe")

    def diffm(self, diffcol: str, name: str="Diff", default=nan):
        """
        Add a diff column to the main dataframe: calculate the
        diff from the column mean

        :param diffcol: column to diff from
        :type diffcol: str
        :param name: diff column name, defaults to "Diff"
        :param name: str, optional
        :param default: column default value, defaults to nan
        :param default: optional

        :example: ``ds.diffm("Col 1", "New col")``
        """
        try:
            df = self.df.copy()
            mean = self.df[diffcol].mean()
            vals = []
            for _, row in self.df.iterrows():
                num = row[diffcol]
                if num > 0:
                    diff = int(((num - mean) * 100) / mean)
                    vals.append(diff)
                else:
                    vals.append(default)
            self.df = df
            self.add(name, vals)
        except Exception as e:
            self.err(e, self._append, "Can not diff column")
            return
        self.ok("Diff column " + name + " added to the dataframe")

    def diffs(self, col: str, serie: "iterable", name: str="Diff"):
        """
        Add a diff column from a serie. The serie is an iterable
        of the same length than the dataframe

        :param col: column to diff
        :type col: str
        :param serie: serie to diff from
        :type serie: iterable
        :param name: name of the diff col, defaults to "Diff"
        :param name: str, optional

        :example: ``ds.diffs("Col 1", [1, 1, 4], "New col")``
        """
        try:
            d = []
            for i, row in self.df.iterrows():
                v = row[col] - serie[i]
                d.append(v)
            self.df[name] = d
        except Exception as e:
            self.err(e, self._append, "Can not diff column from serie")

    def diffsp(self, col: str, serie: "iterable", name: str="Diff"):
        """
        Add a diff column in percentage from a serie. The serie is 
        an iterable of the same length than the dataframe

        :param col: column to diff
        :type col: str
        :param serie: serie to diff from
        :type serie: iterable
        :param name: name of the diff col, defaults to "Diff"
        :param name: str, optional

        :example: ``ds.diffp("Col 1", [1, 1, 4], "New col")``
        """
        try:
            d = []
            for i, row in self.df.iterrows():
                v = (row[col]*100) / serie[i]
                d.append(v)
            self.df[name] = d
        except Exception as e:
            self.err(e, self._append, "Can not diff column from serie")

    def gmean_(self, col: str, index_col: bool=True) -> "Ds":
        """
        Group by and mean column

        :param col: column to group
        :type col: str
        :param index_col: 
        :type index_col: bool
        :return: a dataswim instance
        :rtype: Ds

        :example: ``ds2 = ds.gmean("Col 1")``
        """
        try:
            df = self.df.copy()
            df = df.groupby([col]).mean()
            if index_col is True:
                df[col] = df.index.values
            return self._duplicate_(df)
        except Exception as e:
            self.err(e, self.gmean_, "Can not meansum column")

    def gsum_(self, col: str, index_col: bool=True) -> "Ds":
        """
        Group by and sum column

        :param col: column to group
        :type col: str
        :param index_col: 
        :type index_col: bool
        :return: a dataswim instance
        :rtype: Ds

        :example: ``ds2 = ds.gsum("Col 1")``
        """
        try:
            df = self.df.copy()
            df = df.groupby([col]).sum()
            if index_col is True:
                df[col] = df.index.values
            return self._duplicate_(df)
        except Exception as e:
            self.err(e, self.gsum_, "Can not groupsum column")

    def ratio(self, col: str, ratio_col: str="Ratio"):
        """
        Add a column whith the percentages ratio from a column

        :param col: column to calculate ratio from
        :type col: str
        :param ratio_col: new ratio column name, defaults to "Ratio"
        :param ratio_col: str, optional

        :example: ``ds2 = ds.ratio("Col 1")``
        """
        try:
            df = self.df.copy()
            df[ratio_col] = df[[col]].apply(
                lambda x: 100 * x / float(x.sum()))
            self.df = df
        except Exception as e:
            self.err(e, self.ratio, "Can not calculate ratio")



# @PydevCodeAnalysisIgnore
import holoviews as hv
from .bokeh import Bokeh
from .altair import Altair
from .chartjs import Chartjs
from .seaborn import Seaborn
from .colors import Colors


class Plot(Bokeh, Chartjs, Seaborn, Altair, Colors):
	"""
	Class to handle charts
	"""

	def __init__(self, df=None):
		"""
		Initialize
		"""
		self.df = df
		self.x = None
		self.err = None
		self.y = None
		self.chart_obj = None
		self.chart_opts = dict(width=880)
		self.chart_style = {}
		self.label = None
		self.engine = "bokeh"

	def chart(self, x=None, y=None, chart_type=None, opts=None,
			  style=None, label=None, options={}, **kwargs):
		"""
		Get a chart
		"""
		try:
			self.chart_obj = self._chart(x, y, chart_type, opts, style, label,
										 options=options, **kwargs)
		except Exception as e:
			self.err(e, self.chart, "Can not create chart")

	def chart_(self, x=None, y=None, chart_type=None, opts=None, style=None,
			   label=None, options={}, **kwargs):
		"""
		Get a chart
		"""
		try:
			return self._chart(x, y, chart_type, opts, style, label,
							   options=options, **kwargs)
		except Exception as e:
			self.err(e, self.chart, "Can not create chart")

	def _chart(self, x, y, chart_type, opts, style, label, options, **kwargs):
		"""
		Initialize chart options
		"""
		if opts is not None:
			self.chart_opts = opts
		if style is not None:
			self.chart_style = style
		if label is not None:
			self.label = label
		self.x = x
		self.y = y
		if chart_type is None:
			return
		try:
			chart_obj = self._get_chart(chart_type, x, y, style=style,
										opts=opts, label=label,
										options=options, **kwargs)
			return chart_obj
		except Exception as e:
			self.err(e)

	def bar_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a bar chart
		"""
		try:
			return self._get_chart("bar", style=style, opts=opts, label=label,
								   options=options)
		except Exception as e:
			self.err(e, self.bar_, "Can not draw bar chart")

	def sbar_(self, stack_index=None, label=None, style=None, opts=None,
			  options={}):
		"""
		Get a stacked bar chart
		"""
		self.opts(dict(stack_index=stack_index, color_index=stack_index))
		try:
			if stack_index is None:
				self.err(self.sbar_, "Please provide a stack index parameter")
			options["stack_index"] = stack_index
			return self._get_chart("bar", style=style, opts=opts, label=label,
								   options=options)
		except Exception as e:
			self.err(e, self.sbar_, "Can not draw stacked bar chart")

	def quants_(self, inf, sup, chart_type="point",
				color="green"):
		"""
		Draw a chart to visualize quantiles
		"""
		try:
			return self._bokeh_quants(inf, sup, chart_type, color)
		except Exception as e:
			self.err(e, self.quants_, "Can not draw quantile chart")

	def line_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a line chart
		"""
		try:
			return self._get_chart("line", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.line_, "Can not draw line chart")

	def hline_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a mean line chart
		"""
		try:
			return self._get_chart("hline", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.line_, "Can not draw mean line chart")

	def sline_(self, window_size=5,
			   y_label="Moving average", chart_label=None):
		"""
		Get a moving average curve chart to smooth between points
		"""
		options = dict(window_size=window_size, y_label=y_label)
		try:
			return self._get_chart("sline", style=self.chart_style,
								   opts=self.chart_opts, label=chart_label,
								   options=options)
		except Exception as e:
			self.err(e, self.sline_, "Can not draw smooth curve chart")

	def area_(self, label=None, style=None, opts=None, options={}):
		"""
		Get an area chart
		"""
		try:
			return self._get_chart("area", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.area_, "Can not draw area chart")

	def sarea_(self, col, x=None, y=None, rsum=None, rmean=None):
		"""
		Get an stacked area chart
		"""
		try:
			charts = self._multiseries(col, x, y, "area", rsum, rmean)
			return hv.Area.stack(charts)
		except Exception as e:
			self.err(e, self.sarea_, "Can not draw stacked area chart")

	def hist_(self, label=None, style=None, opts=None, options={}):
		"""
		Get an historiogram chart
		"""
		try:
			return self._get_chart("hist", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.hist_, "Can not draw historiogram")

	def errorbar_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a point chart
		"""
		try:
			return self._get_chart("errorBar", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.errorbar_, "Can not draw errorbar chart")

	def point_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a point chart
		"""
		try:
			return self._get_chart("point", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.point_, "Can not draw point chart")

	def circle_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a circle chart
		"""
		try:
			return self._get_chart("circle", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.circle_, "Can not draw circle chart")

	def square_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a square chart
		"""
		try:
			return self._get_chart("square", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.square_, "Can not draw square chart")

	def tick_(self, label=None, style=None, opts=None, options={}):
		"""
		Get an tick chart
		"""
		try:
			self._get_chart("tick", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, "Can not draw tick chart")

	def rule_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a rule chart
		"""
		try:
			return self._get_chart("rule", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.rule_, "Can not draw rule chart")

	def heatmap_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a heatmap chart
		"""
		try:
			return self._get_chart("heatmap", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.heatmap_, "Can not draw heatmap")

	def lreg_(self, label=None, style=None, opts=None, options={}):
		"""
		Get a linear regression chart
		"""
		try:
			return self._get_chart("lreg", style=style, opts=opts,
								   label=label, options=options)
		except Exception as e:
			self.err(e, self.rule_, "Can not draw linear regression chart")

	def line_point_(self, label=None, style=None, opts=None, options={},
					colors={"line": "orange", "point": "#30A2DA"}):
		"""
		Get a line and point chart
		"""
		try:
			if style is None:
				style = self.chart_style
				if "size" not in style:
					style["size"] = 10
			style["color"] = colors["line"]
			c = self._get_chart("line", style=style, opts=opts,
								label=label, options=options)
			style["color"] = colors["point"]
			c2 = self._get_chart("point", style=style, opts=opts,
								 label=label, options=options)
			return c * c2
		except Exception as e:
			self.err(e, self.line_point_, "Can not draw line_point chart")

	def mpoint_(self, col, x=None, y=None, rsum=None, rmean=None):
		"""
		Splits a column into multiple series based on the column's
		unique values. Then visualize theses series in a chart.
		Parameters: column to split, x axis column, y axis column
		Optional: rsum="1D" to resample and sum data an rmean="1D"
		to mean the data
		"""
		return self._multiseries(col, x, y, "point", rsum, rmean)

	def mline_(self, col, x=None, y=None, rsum=None, rmean=None):
		"""
		Splits a column into multiple series based on the column's
		unique values. Then visualize theses series in a chart.
		Parameters: column to split, x axis column, y axis column
		Optional: rsum="1D" to resample and sum data an rmean="1D"
		to mean the data
		"""
		return self._multiseries(col, x, y, "line", rsum, rmean)

	def mbar_(self, col, x=None, y=None, rsum=None, rmean=None):
		"""
		Splits a column into multiple series based on the column's
		unique values. Then visualize theses series in a chart.
		Parameters: column to split, x axis column, y axis column
		Optional: rsum="1D" to resample and sum data an rmean="1D"
		to mean the data
		"""
		return self._multiseries(col, x, y, "bar", rsum, rmean)

	def mline_point_(self, col, x=None, y=None, rsum=None, rmean=None):
		"""
		Splits a column into multiple series based on the column's
		unique values. Then visualize theses series in a chart.
		Parameters: column to split, x axis column, y axis column
		Optional: rsum="1D" to resample and sum data an rmean="1D"
		to mean the data
		"""
		line = self._multiseries(col, x, y, "line", rsum, rmean)
		point = self._multiseries(col, x, y, "point", rsum, rmean)
		return line * point

	def arrow_(self, xloc, yloc, text, orientation="v", arrowstyle='->'):
		"""
		Returns an arrow for a chart. Params: the text, xloc and yloc are
		coordinates to position the arrow. Orientation is the way to display
		the arrow: possible values are ``[<, ^, >, v]``. Arrow style is the
		graphic style of the arrow: 
		possible values: ``[-, ->, -[, -|>, <->, <|-|>]``
		"""
		try:
			arrow = hv.Arrow(
				xloc,
				yloc,
				text,
				orientation,
				arrowstyle=arrowstyle)
			return arrow
		except Exception as e:
			self.err(e, self.arrow_, "Can not draw arrow chart")

	def _multiseries(self, col, x, y, ctype, rsum, rmean):
		"""
		Chart multiple series from a column distinct values
		"""
		self.autoprint = False
		x, y = self._check_fields(x, y)
		chart = None
		series = self.split_(col)
		for key in series:
			instance = series[key]
			if rsum is not None:
				instance.rsum(rsum, index_col=x)
			if rmean is not None:
				instance.rmean(rmean, index_col=x)
			instance.chart(x, y)
			self.scolor()
			c = None
			label = str(key)
			if ctype == "point":
				c = instance.point_(label)
			if ctype == "line":
				instance.zero_nan(y)
				c = instance.line_(label)
			if ctype == "bar":
				c = instance.bar_(label)
			if ctype == "area":
				c = instance.area_(label)
			if c is None:
				self.warning("Chart type " + ctype + 
							 " not supported, aborting")
				return
			if chart is None:
				chart = c
			else:
				chart = chart * c
		return chart

	def density_(self, label=None, style=None, opts=None):
		"""
		Get a Seaborn density chart
		"""
		try:
			return self._get_chart("density", style=style, opts=opts, label=label)
		except Exception as e:
			self.err(e, "Can not draw density chart")
			
	def dlinear_(self, label=None, style=None, opts=None):
		"""
		Get a Seaborn linear + distribution chart
		"""
		try:
			return self._get_chart("dlinear", style=style, opts=opts, label=label)
		except Exception as e:
			self.err(e, "Can not draw dlinear chart")
			
	def distrib_(self, label=None, style=None, opts=None):
		"""
		Get a Seaborn distribution chart
		"""
		try:
			return self._get_chart("distribution", style=style, opts=opts, label=label)
		except Exception as e:
			self.err(e, "Can not draw distrinution chart")

	def text_(self, label=None, style=None, opts=None):
		"""
		Get an Altair text marks chart
		"""
		try:
			return self._get_chart("text", style=style, opts=opts, label=label)
		except Exception as e:
			self.err(e, "Can not draw text marks chart")

	def line_num_(self, label=None, style=None, opts=None):
		"""
		Get an Altair line + number marks chart
		"""
		try:
			return self._get_chart("line_num", style=style, opts=opts, label=label)
		except Exception as e:
			self.err(e, "Can not draw line and numbers chart")

	def bar_num_(self, label=None, style=None, opts=None):
		"""
		Get an Altair bar + number marks chart
		"""
		try:
			return self._get_chart("bar_num", style=style, opts=opts, label=label)
		except Exception as e:
			self.err(e, "Can not draw line and numbers chart")
			
	def point_num_(self, label=None, style=None, opts=None):
		"""
		Get an Altair point + number marks chart
		"""
		try:
			return self._get_chart("point_num", style=style, opts=opts, label=label)
		except Exception as e:
			self.err(e, "Can not draw point and numbers chart")

	def opt(self, name, value):
		"""
		Add or update one option
		"""
		self.chart_opts[name] = value

	def style(self, name, value):
		"""
		Add or update one style
		"""
		self.chart_style[name] = value

	def opts(self, dictobj):
		"""
		Add or update options
		"""
		for k in dictobj:
			self.chart_opts[k] = dictobj[k]

	def styles(self, dictobj):
		"""
		Add or update styles
		"""
		for k in dictobj:
			self.chart_style[k] = dictobj[k]

	def ropt(self, name):
		"""
		Remove one option
		"""
		try:
			del self.chart_opts[name]
		except KeyError:
			self.warning("Option " + name + " is not set")
		except:
			self.err("Can not remove option " + name)

	def rstyle(self, name):
		"""
		Remove one style
		"""
		try:
			del self.chart_style[name]
		except KeyError:
			self.warning("Style " + name + " is not set")
		except:
			self.err("Can not remove style " + name)

	def ropts(self):
		"""
		Reset the chart options
		"""
		self.chart_opts = {}

	def rstyles(self):
		"""
		Reset the chart options
		"""
		self.chart_style = {}

	def defaults(self):
		"""
		Reset the chart options and style to defaults
		"""
		self.chart_style = {}
		self.chart_opts = {}
		self.style("color", "#30A2DA")
		self.width(900)
		self.height(250)

	def color(self, val):
		"""
		Change the chart's color
		"""
		self.style("color", val)

	def rcolor(self):
		"""
		Reset the color to the base color
		"""
		self.style("color", "#30A2DA")

	def width(self, val):
		"""
		Change the chart's width
		"""
		self.opts(dict(width=val))

	def height(self, val):
		"""
		Change the chart's height
		"""
		self.opts(dict(height=val))

	def size(self, val):
		"""
		Change the chart's point size
		"""
		self.chart_style["size"] = val

	def _check_fields(self, x, y):
		"""
		Check x and y fields parameters and initialize
		"""
		if x is None:
			if self.x is None:
				self.err(
					self._check_fields,
					"X field is not set: please specify a parameter")
				return
			x = self.x
		if y is None:
			if self.y is None:
				self.err(
					self._check_fields,
					"Y field is not set: please specify a parameter")
				return
			y = self.y
		return x, y

	def _get_chart(self, chart_type, x=None, y=None, style=None, opts=None,
				   label=None, options={}, **kwargs):
		"""
		Get a full chart object
		"""
		sbcharts = ["density", "distribution", "dlinear"]
		acharts = ["tick", "circle", "text", "line_num", "bar_num"]
		if chart_type in sbcharts:
			self._set_seaborn_engine()
		if chart_type in acharts:
			self._set_altair_engine()
		if chart_type != "sline":
			x, y = self._check_fields(x, y)
		if opts is None:
			opts = self.chart_opts
		if style is None:
			style = self.chart_style
		if self.engine == "bokeh":
			func = self._get_bokeh_chart
		elif self.engine == "altair":
			func = self._get_altair_chart
		elif self.engine == "chartjs":
			func = self._get_chartjs_chart
		elif self.engine == "seaborn":
			func = self._get_seaborn_chart
		else:
			self.err("Engine " + self.engine + " unknown")
			return
		try:
			chart = func(
				x, y, chart_type, label, opts, style,
				options=options, **kwargs)
			return chart
		except Exception as e:
			self.err(e)

	def _check_defaults(self, x_only=True):
		"""
		Checks if charts defaults are set
		"""
		if self.x is None:
			self.err(self._check_defaults,
					 "X field is not set: please specify a parameter")
			return
		if x_only is True:
			return
		if self.y is None:
			self.err(self._check_defaults,
					 "Y field is not set: please specify a parameter")
			return



# @PydevCodeAnalysisIgnore
import holoviews as hv
import numpy as np
from bokeh.embed import components
from holoviews.core.data.interface import DataError

bokeh_renderer = hv.renderer('bokeh')


class Bokeh():
    """
    A class to handle charts with the Bokeh library
    """

    def __init__(self, df=None, db=None):
        """
        Initialize
        """
        self.df = df
        self.x_field = None
        self.y_field = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = None
        self.label = None
        self.engine = "bokeh"

    def ndlayout_(self, dataset, kdims, cols=3):
        """
        Create a Holoview NdLayout from a dictionnary of chart objects
        """
        try:
            return hv.NdLayout(dataset, kdims=kdims).cols(cols)
        except Exception as e:
            self.err(e, self.layout_, "Can not create layout")

    def layout_(self, chart_objs, cols=3):
        """
        Returns a Holoview Layout from chart objects
        """
        try:
            return hv.Layout(chart_objs).cols(cols)
        except Exception as e:
            self.err(e, self.layout_, "Can not build layout")

    def bokeh_header_(self):
        """
        Returns html script tags for Bokeh
        """
        header = """
        <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/bokeh/0.12.11/bokeh.css" />
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/bokeh/0.12.11/bokeh.min.js"></script>
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/bokeh/0.12.11/bokeh.min.js.map"></script>
        <script type="text/javascript">
            Bokeh.set_log_level("info");
        </script>
        """
        return header

    def _hline_bokeh_(self, col):
        """
        Returns an horizontal line from a column mean value
        """
        c = hv.HLine(self.df[col].mean())
        return c

    """def spikes_(self, col, label=None):
        ""
        Returns spikes chart from a column mean value
        ""
        try:
            chart = hv.Spikes(list(self.df[col]), self.x, self.y)
            return chart
        except Exception as e:
            self.err(e, self.spikes_, "Can not fraw spikes chart")"""

    def _sline_bokeh(self, window_size, y_label):
        """
        Returns a chart with a smooth line from a serie
        """
        try:
            ds2 = self.clone_()
            window = np.ones(int(window_size)) / float(window_size)
            ds2.df[y_label] = np.convolve(self.df[self.y], window, 'same')
            ds2.chart(self.x, y_label)
            return ds2.line_()
        except Exception as e:
            self.err(e, self._sline_bokeh, "Can not draw smooth line chart")

    def _lreg_bokeh(self, **kwargs):
        """
        Returns a Bokeh linear regression line
        """
        try:
            ds2 = self._duplicate_()
            ds2.timestamps(ds2.x)
            ds2.lreg("Timestamps", ds2.y)
            ds2.drop(ds2.y)
            ds2.df = ds2.df.rename(columns={'Regression': ds2.y})
            if "date_format" in self.chart_style:
                ds2.date("Date", format=self.chart_style["date_format"])
            c = ds2.line_()
            return c
        except Exception as e:
            self.err(e, "Can not draw linear regression chart")

    def _bokeh_quants(self, inf, sup, chart_type, color):
        """
        Draw a chart to visualize quantiles
        """
        try:
            ds2 = self._duplicate_()
            qi = ds2.df[ds2.y].quantile(inf)
            qs = ds2.df[ds2.y].quantile(sup)
            ds2.add("sup", qs)
            ds2.add("inf", qi)
            ds2.chart(ds2.x, ds2.y)
            if chart_type == "point":
                c = ds2.point_(opts=self.chart_opts, style=self.chart_style)
            elif chart_type == "line_point":
                c = ds2.line_point_(opts=self.chart_opts, style=self.chart_style)
            else:
                c = ds2.line_(opts=self.chart_opts, style=self.chart_style)
            ds2.color(color)
            ds2.chart(ds2.x, "sup")
            c2 = ds2.line_()
            ds2.chart(ds2.x, "inf")
            c3 = ds2.line_()
            ds2.rcolor()
            return c * c2 * c3
        except Exception as e:
            self.err(e, "Can not draw quantile chart")

    def _get_bokeh_chart(self, x_field, y_field, chart_type,
                         label, opts, style, options={}, **kwargs):
        """
        Get a Bokeh chart object
        """
        if isinstance(x_field, list):
            kdims = x_field
        else:
            kdims = [x_field]
        if isinstance(y_field, list):
            vdims = y_field
        else:
            vdims = [y_field]
        args = kwargs
        args["data"] = self.df
        args["kdims"] = kdims
        args["vdims"] = vdims
        if label is not None:
            args["label"] = label
        else:
            if self.label is not None:
                args["label"] = self.label
        chart = None
        try:
            if chart_type == "line":
                chart = hv.Curve(**args)
            if chart_type == "hline":
                chart = self._hline_bokeh_(y_field)
            elif chart_type == "point":
                chart = hv.Scatter(**args)
            elif chart_type == "area":
                chart = hv.Area(**args)
            elif chart_type == "bar":
                chart = hv.Bars(**args)
            elif chart_type == "hist":
                chart = hv.Histogram(**args)
            elif chart_type == "errorBar":
                chart = hv.ErrorBars(**args)
            elif chart_type == "heatmap":
                chart = hv.HeatMap(**args)
            elif chart_type == "lreg":
                chart = self._lreg_bokeh(**args)
            elif chart_type == "sline":
                window_size, y_label = options["window_size"], options["y_label"]
                chart = self._sline_bokeh(window_size, y_label)
            if chart is None:
                self.err("Chart type " + chart_type + 
                         " unknown", self._get_bokeh_chart)
                return
            endchart = chart(plot=opts, style=style)
            return endchart
        except DataError as e:
            msg = "Column not found in " + x_field + " and " + y_field
            self.err(e, self._get_bokeh_chart, msg)
        except Exception as e:
            self.err(e)

    def _get_bokeh_html(self, chart_obj):
        """
        Get the html for a Bokeh chart
        """
        global bokeh_renderer
        try:
            renderer = bokeh_renderer
            p = renderer.get_plot(chart_obj).state
            script, div = components(p)
            return script + "\n" + div

        except Exception as e:
            self.err(e, self._get_bokeh_html,
                     "Can not get html from the Bokeh rendering engine")



# -*- coding: utf-8 -*-
import datetime
import pandas as pd
import numpy as np
from pandas.testing import assert_series_equal, assert_frame_equal
from dataswim.tests.base import BaseDsTest
from dataswim import Ds

ds = Ds()

ds.errs_traceback = False


class TestDsDataSelect(BaseDsTest):

    def test_first(self):
        df1 = pd.DataFrame([["one", "three"], ["two", "four"]])
        ds.df = df1
        first = ds.first_()
        res = pd.Series(["one", "three"], name=0)
        assert_series_equal(first, res)
        ds.df = None
        _ = ds.first_()
        self.assertRaises(AttributeError)

    def test_unique(self):
        df1 = pd.DataFrame([["one", "one", "three"], ["two", "two", "four"]])
        ds.df = df1
        data = ds.unique_(0)
        res = ["one", "two"]
        self.assertEqual(data, res)
        ds.df = None
        _ = ds.unique_(0)
        self.assertRaises(AttributeError)

    """def test_range(self):
        dates = pd.date_range('20130101', periods=6)
        vals = np.random.randn(6)
        df1 = pd.DataFrame(vals, index=dates)
        ds.df = df1
        ds.vrange(days=2)
        dates2 = pd.date_range('20130101', periods=3)
        df2 = pd.DataFrame(vals[:3], index=dates2)
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.range_(days=2)
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        ds.vrange(days=2)
        self.assertRaises(AttributeError)
        _ = ds.range_(days=2)
        self.assertRaises(AttributeError)"""

    def test_daterange(self):
        dates = dates = pd.date_range('20130101', periods=6)
        df1 = pd.DataFrame(dates, columns=["one"])
        ds.df = df1
        ds.daterange("one", "2013-01-02", "+", days=2)
        dates2 = pd.date_range('20130102', periods=3)
        df2 = pd.DataFrame(dates2, columns=["one"], index=[1, 2, 3])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.daterange_("one", "2013-01-02", "+", days=2)
        assert_frame_equal(ds2.df, df2)
        ds.df = df1
        ds.daterange("one", "2013-01-06", "-", days=2)
        dates3 = pd.date_range('20130104', periods=3)
        df3 = pd.DataFrame(dates3, columns=["one"], index=[3, 4, 5])
        assert_frame_equal(ds.df, df3)
        ds.df = None
        _ = ds.daterange_("one", "2013-01-02", "+", days=2)
        self.assertRaises(AttributeError)
        ds.daterange("one", "2013-01-02", "+", days=2)
        self.assertRaises(AttributeError)

    def test_subset(self):
        df1 = pd.DataFrame([["one", "one"], ["two", "two"],
                            ["three", "three"]])
        ds.df = df1
        ds.subset(2)
        df2 = pd.DataFrame([["one", "one"], ["two", "two"]])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.subset_(0, 2)
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        ds.subset(2)
        self.assertRaises(AttributeError)
        _ = ds.subset_(2)
        self.assertRaises(AttributeError)

    """def test_nowrange(self):
        vdates = list(pd.date_range(datetime.datetime.now(), periods=6))
        dates = []
        for d in vdates:
            dates.append([d])
        vals = np.random.randn(6)
        df1 = pd.DataFrame(dates, columns=["one"])
        ds.df = df1
        ds.nowrange("one", 3)
        dates2 = dates[:3]
        df2 = pd.DataFrame(dates2, columns=["one"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.nowrange_("one", 3)
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        ds.nowrange("one", 3)
        self.assertRaises(AttributeError)
        _ = ds.nowrange_("one", 3)
        self.assertRaises(AttributeError)"""

    def test_limit(self):
        df1 = pd.DataFrame([["one", "three"], ["two", "four"]])
        ds.df = df1
        ds2 = ds.limit_(1)
        df2 = pd.DataFrame([["one", "three"]])
        assert_frame_equal(ds2.df, df2)
        ds.df = df1
        ds.limit(1)
        assert_frame_equal(ds.df, df2)
        ds.df = None
        _ = ds.limit_()
        self.assertRaises(AttributeError)
        ds.limit()
        self.assertRaises(AttributeError)



from collections import Counter
import pandas as pd
import nltk


class Text():
    """
    Class to process text
    """

    def flat_(self, col, nums=True):
        """
        Returns a flat representation of a column's values
        """
        try:
            res = ""
            i = 0
            vals = self.df[col].tolist()
            for el in vals:
                if nums is True:
                    res = res + str(i) + " " + el
                else:
                    res = res + el
                if i < (len(vals)-1):
                    res += " "
                i += 1
            return res
        except Exception as e:
            self.err(e, "Can not flat " + col)

    def mfw_(self, col, sw_lang="english", limit=100):
        """
        Returns a Dataswim instance with the most frequent words in a
        column exluding the most common stop words
        """
        df = self._mfw(col, sw_lang, limit)
        if df is None:
            self.err("Can not find most frequent words")
            return
        return self._duplicate_(df)

    def _mfw(self, col, sw_lang, limit):
        try:
            stopwords = nltk.corpus.stopwords.words(sw_lang)
            RE_stopwords = r'\b(?:{})\b'.format('|'.join(stopwords))
            # replace '|'-->' ' and drop all stopwords
            words = (self.df[col]
                     .str.lower()
                     .replace([r'\|', RE_stopwords], [' ', ''], regex=True)
                     .str.cat(sep=' ')
                     .split()
                     )
            df = pd.DataFrame(Counter(words).most_common(limit),
                              columns=['Word', 'Frequency']).set_index('Word')
        except Exception as e:
            self.err(e, "Can not finc most frequent words")
            return
        return df



# @PydevCodeAnalysisIgnore
import arrow
import pandas as pd
import numpy as np
from ..errors import Error


class Select(Error):
    """
    Class to select data
    """

    def first_(self):
        """
        Select the first row
        """
        try:
            val = self.df.iloc[0]
            return val
        except Exception as e:
            self.err(e, "Can not select first row")

    def limit(self, r=5):
        """
        Limit selection to a range in the main dataframe
        """
        try:
            self.df = self.df[:r]
        except Exception as e:
            self.err(e, "Can not limit data")

    def limit_(self, r=5):
        """
        Returns a DataSwim instance with limited selection
        """
        try:
            return self._duplicate_(self.df[:r])
        except Exception as e:
            self.err(e, "Can not limit data")

    def unique_(self, col):
        """
        Returns unique values in a column
        """
        try:
            df = self.df.drop_duplicates(subset=[col], inplace=False)
            return list(df[col])
        except Exception as e:
            self.err(e, "Can not select unique data")

    def wunique_(self, col):
        """
        Weight unique values: returns a dataframe with a count
        of unique values
        """
        try:
            s = pd.value_counts(self.df[col].values)
            df = pd.DataFrame(s, columns=["Number"])
            return df
        except Exception as e:
            self.err(e, "Can not weight unique data")

    """def ranged(self, end_date, start_date=None):
        ""
        Limit the data in a time range between two dates. The
        dataframe must have a valid date index
        ""
        try:
            if start_date is None:
                start_date = self.df.index.get_values()[0]
            self.df = self.df.loc[start_date: end_date]
        except Exception as e:
            self.err("Can not select range data", e)

    def rangeo(self, **kwargs):
        ""
        Limit the data in a time range between an end date and an offset.
        The dataframe must have a valid date index. The offset must be
        provided: ex: months=3. An end_date can be provided optionally. If
        not provided the last index of the dataframe will be used as end date
        ""
        try:
            if "end_date" not in kwargs:
                end_date = self.df.index.get_values()[len(self.df.index)-1]
            else:
                end_date = kwargs["end_date"]
            if "years" in kwargs:
                start_date = end_date - pd.DateOffset(years=kwargs["years"])
            if "months" in kwargs:
                print("M", kwargs["months"])
                start_date = end_date - pd.DateOffset(months=kwargs["months"])
            if "days" in kwargs:
                start_date = end_date - pd.DateOffset(days=kwargs["days"])
            if "hours" in kwargs:
                start_date = end_date - pd.DateOffset(hours=kwargs["hours"])
            if "minutes" in kwargs:
                start_date = end_date - \
                    pd.DateOffset(minutes=kwargs["minutes"])
            if "seconds" in kwargs:
                start_date = end_date - \
                    pd.DateOffset(seconds=kwargs["seconds"])
            self.df = self.df.loc[start_date: end_date]
        except Exception as e:
            self.err("Can not select range data", e)"""

    def nowrange(self, col, timeframe):
        """
        Set the main dataframe with rows within a date range from now
        ex: ds.nowrange("Date", "3D") for a 3 days range. Units are: S,
        H, D, W, M, Y
        """
        df = self._nowrange(col, timeframe)
        if df is None:
            self.err("Can not select range data from now")
            return
        self.df = df

    def nowrange_(self, col, timeframe):
        """
        Returns a Dataswim instance with rows within a date range from now
        ex: ds.nowrange("Date", "3D") for a 3 days range. Units are: S,
        H, D, W, M, Y
        """
        df = self._nowrange(col, timeframe)
        if df is None:
            self.err("Can not select range data from now")
            return
        return self._duplicate_(df)

    def _nowrange(self, col, timeframe):
        try:
            # df = self.df[self.df[col].dt.date < datetime.now().date(
            # ) + pd.to_timedelta(interval, unit=unit)]
            unit = timeframe[-1:]
            i = int(timeframe[0:(len(timeframe) - 1)])
            interval = int(np.negative(i))
            if unit == "S":
                date = arrow.utcnow().shift(seconds=interval)
            elif unit == "m":
                date = arrow.utcnow().shift(minutes=interval)
            elif unit == "H":
                date = arrow.utcnow().shift(hours=interval)
            elif unit == "D":
                date = arrow.utcnow().shift(days=interval)
            elif unit == "W":
                date = arrow.utcnow().shift(weeks=interval)
            elif unit == "M":
                date = arrow.utcnow().shift(months=interval)
            elif unit == "Y":
                date = arrow.utcnow().shift(years=interval)
            else:
                self.err("Wrong unit " + unit)
                return
            df = self.df.copy()
            df = df[df[col] > date.datetime]
            return df
        except Exception as e:
            self.err(e, "Can not select range data from now")

    def daterange(self, datecol, date_start, op, **args):
        """
        Returns rows in a date range
        """
        df = self._daterange(datecol, date_start, op, **args)
        if df is None:
            self.err("Can not select date range data")
        self.df = df

    def daterange_(self, datecol, date_start, op, **args):
        """
        Returns a DataSwim instance with rows in a date range
        """
        df = self._daterange(datecol, date_start, op, **args)
        if df is None:
            self.err("Can not select date range data")
        return self._duplicate_(df)

    def _daterange(self, datecol, date_start, op, **args):
        try:
            idate = pd.Timestamp(date_start)
            self.df[datecol] = pd.to_datetime(self.df[datecol])
            if op == "+":
                end_date = idate + pd.DateOffset(**args)
                start_date = idate
            elif op == "-":
                start_date = idate - pd.DateOffset(**args)
                end_date = idate
            mask = (self.df[datecol] >= start_date) & (
                self.df[datecol] <= end_date)
            df = self.df.loc[mask]
            return df
        except Exception as e:
            self.err(e, self._daterange, "Can not select date range data")

    def subset(self, *args):
        """
        Set the main dataframe to a subset based in positions
        Select a subset of the main dataframe based on position:
        ex: ds.subset(0,10) or ds.subset(10) is equivalent: it starts
        at the first row if only one argument is provided
        """
        df = self._subset(*args)
        if df is None:
            self.err("Can get subset of data")
            return
        self.df = df

    def subset_(self, *args):
        """
        Returns a Dataswim instance with a subset data based in positions
        Select a subset of the main dataframe based on position:
        ex: ds.subset(0,10) or ds.subset(10) is equivalent: it starts
        at the first row if only one argument is provided
        """
        df = self._subset(*args)
        if df is None:
            self.err("Can not get subset of data")
            return
        return self._duplicate_(df)

    def _subset(self, *args):
        try:
            start = 0
            end = args[0]
            if len(args) > 1:
                start = args[0]
                end = args[1]
            return self.df.iloc[start: end]
        except Exception as e:
            self.err(e, "Can not select data")



import pandas as pd
from ..copy import Copy


class Dataframe(Copy):
    """
    Class to transform the dataframe
    """

    def concat(self, *dss, **kwargs):
        """
        Concatenate dataswim instances from and
                set it to the main dataframe

        :param dss: dataswim instances to concatenate
        :type dss: Ds
        :param kwargs: keyword arguments for ``pd.concat``


        """
        try:
            df = pd.DataFrame()
            for dsx in dss:
                df = pd.concat([df, dsx.df], **kwargs)
            self.df = df
        except Exception as e:
            self.err(e, "Can not concatenate data")

    def concat_(self, *dss, **kwargs):
        """
        Concatenate dataswim instances and
                return a new Ds instance

        :param dss: dataswim instances to concatenate
        :type dss: Ds
        :param kwargs: keyword arguments for ``pd.concat``
        :rtype: Ds

        """
        try:
            df = pd.DataFrame()
            for dsx in dss:
                df = pd.concat([df, dsx.df], **kwargs)
            return self.new_(df=df)
        except Exception as e:
            self.err(e, "Can not concatenate data")

    def split_(self, col: str) -> "list(Ds)":
        """
        Split the main dataframe according to a column's unique values and
        return a dict of dataswim instances

        :return: list of dataswim instances
        :rtype: list(Ds)

        :example: ``dss = ds.slit_("Col 1")``
        """
        try:
            dss = {}
            unique = self.df[col].unique()
            for key in unique:
                df2 = self.df.loc[self.df[col] == key]
                ds2 = self._duplicate_(df2)
                dss[key] = ds2
            return dss
        except Exception as e:
            self.err(e, "Can not split dataframe")

    def merge(self, df: pd.DataFrame, on: str, how: str="outer", **kwargs):
        """
        Set the main dataframe from the current dataframe and the passed
        dataframe

        :param df: the pandas dataframe to merge
        :type df: pd.DataFrame
        :param on: param for ``pd.merge``
        :type on: str
        :param how: param for ``pd.merge``, defaults to "outer"
        :type how: str, optional
        :param kwargs: keyword arguments for ``pd.merge``
        """
        try:
            df = pd.merge(self.df, df, on=on, how=how, **kwargs)
            self.df = df
        except Exception as e:
            self.err(e, self.merge, "Can not merge dataframes")



# @PydevCodeAnalysisIgnore
from .columns import Columns
from .values import Values
from .resample import Resample
from .dataframe import Dataframe
from .calculations import Calculations


class Transform(Columns, Values, Resample, Dataframe, Calculations):
    """
    Class to transform data
    """

    pass
